import os
import os.path
import tempfile
from typing import Dict, List, Optional, Set

import libdnf5 as dnf5
from libdnf5.base import GoalProblem_EXCLUDED as EXCLUDED
from libdnf5.base import GoalProblem_NO_PROBLEM as NO_PROBLEM
from libdnf5.base import GoalProblem_NOT_FOUND as NOT_FOUND
from libdnf5.common import QueryCmp_CONTAINS as CONTAINS
from libdnf5.common import QueryCmp_EQ as EQ
from libdnf5.common import QueryCmp_GLOB as GLOB

import osbuild.solver.model as model
from osbuild.solver import SolverBase, modify_rootdir_path, read_keys
from osbuild.solver.exceptions import DepsolveError, MarkingError, NoReposError, RepoError
from osbuild.solver.request import DepsolveCmdArgs, SearchCmdArgs, SolverConfig
from osbuild.util.sbom.dnf5 import dnf_pkgset_to_sbom_pkgset
from osbuild.util.sbom.spdx import sbom_pkgset_to_spdx2_doc

# GoalProblem flags that represent package marking/selection failures (as
# opposed to dependency resolution failures). When the returned problem
# bitmask contains only these flags, we raise MarkingError to be consistent
# with DNF4, which raises MarkingError from install_specs().
MARKING_PROBLEMS = NOT_FOUND | EXCLUDED


def remote_location(package, schemes=("http", "ftp", "file", "https")):
    """Return the remote url where a package rpm may be downloaded from

    This wraps the get_remote_location() function, returning the first
    result or if it cannot find a suitable url it raises a RuntimeError
    """
    urls = package.get_remote_locations(schemes)
    if not urls or len(urls) == 0:
        raise RuntimeError(f"Cannot determine remote location for {package.get_nevra()}")

    return urls[0]


def get_string_option(option):
    # option.get_value() causes an error if it's unset for string values, so check if it's empty first
    if option.empty():
        return None
    return option.get_value()


# XXX - Temporarily lifted from dnf.rpm module  # pylint: disable=fixme
def _invert(dct):
    return {v: k for k in dct for v in dct[k]}


def any_repos_enabled(base):
    """Return true if any repositories are enabled"""
    rq = dnf5.repo.RepoQuery(base)
    return rq.begin() != rq.end()


def _reldep_to_dependency(reldep: dnf5.rpm.Reldep) -> model.Dependency:
    """
    Convert a libdnf5.rpm.Reldep to an RPM Dependency.
    """
    return model.Dependency(reldep.get_name(), reldep.get_relation().strip(), reldep.get_version())


# pylint: disable=too-many-branches
def _dnf_pkg_to_package(pkg: dnf5.rpm.Package) -> model.Package:
    """
    Convert a dnf5.rpm.Package to an RPM Package.
    """
    kwargs = {
        "name": pkg.get_name(),
        "version": pkg.get_version(),
        "release": pkg.get_release(),
        "arch": pkg.get_arch(),
        "epoch": int(pkg.get_epoch()),
        "group": pkg.get_group() or "",
        "download_size": pkg.get_download_size() or 0,
        "install_size": pkg.get_install_size() or 0,
        "license": pkg.get_license() or "",
        "source_rpm": pkg.get_sourcerpm() or "",
        "build_time": pkg.get_build_time() or 0,
        "packager": pkg.get_packager() or "",
        "vendor": pkg.get_vendor() or "",
        "url": pkg.get_url() or "",
        "summary": pkg.get_summary() or "",
        "description": pkg.get_description() or "",
        "provides": sorted([_reldep_to_dependency(p) for p in pkg.get_provides()], key=str),
        "requires": sorted([_reldep_to_dependency(p) for p in pkg.get_requires()], key=str),
        "requires_pre": sorted([_reldep_to_dependency(p) for p in pkg.get_requires_pre()], key=str),
        "conflicts": sorted([_reldep_to_dependency(p) for p in pkg.get_conflicts()], key=str),
        "obsoletes": sorted([_reldep_to_dependency(p) for p in pkg.get_obsoletes()], key=str),
        "regular_requires": sorted([_reldep_to_dependency(p) for p in pkg.get_regular_requires()], key=str),
        "recommends": sorted([_reldep_to_dependency(p) for p in pkg.get_recommends()], key=str),
        "suggests": sorted([_reldep_to_dependency(p) for p in pkg.get_suggests()], key=str),
        "enhances": sorted([_reldep_to_dependency(p) for p in pkg.get_enhances()], key=str),
        "supplements": sorted([_reldep_to_dependency(p) for p in pkg.get_supplements()], key=str),
        "files": list(pkg.get_files()),
        "location": pkg.get_location() or "",
        "remote_locations": list(pkg.get_remote_locations()),
        "repo_id": pkg.get_repo().get_id() or "",
        "reason": pkg.get_reason() or "",
    }
    checksum = pkg.get_checksum()
    if checksum and checksum.get_type() != dnf5.rpm.Checksum.Type_UNKNOWN:
        kwargs["checksum"] = model.Checksum(
            algorithm=checksum.get_type_str(), value=checksum.get_checksum())
    header_checksum = pkg.get_hdr_checksum()
    if header_checksum and header_checksum.get_type() != dnf5.rpm.Checksum.Type_UNKNOWN:
        kwargs["header_checksum"] = model.Checksum(
            algorithm=header_checksum.get_type_str(), value=header_checksum.get_checksum())
    return model.Package(**kwargs)


def _dnf_repo_to_repository(
    repo: dnf5.repo.Repo,
    root_dir: Optional[str],
    request_repo_ids: Set[str],
) -> model.Repository:
    """
    Convert a dnf5.repo.Repo to a Repository.
    """
    repo_cfg = repo.get_config()
    metadata_expire = get_string_option(repo_cfg.get_metadata_expire_option())
    return model.Repository(
        repo_id=repo.get_id(),
        name=repo.get_name(),
        baseurl=list(repo_cfg.get_baseurl_option().get_value()),
        metalink=get_string_option(repo_cfg.get_metalink_option()),
        mirrorlist=get_string_option(repo_cfg.get_mirrorlist_option()),
        gpgcheck=repo_cfg.get_pkg_gpgcheck_option().get_value(),
        repo_gpgcheck=repo_cfg.get_repo_gpgcheck_option().get_value(),
        gpgkey=read_keys(
            repo_cfg.get_gpgkey_option().get_value(),
            root_dir if repo.get_id() not in request_repo_ids else None
        ),
        sslverify=repo_cfg.get_sslverify_option().get_value(),
        sslcacert=get_string_option(repo_cfg.get_sslcacert_option()),
        sslclientkey=get_string_option(repo_cfg.get_sslclientkey_option()),
        sslclientcert=get_string_option(repo_cfg.get_sslclientcert_option()),
        metadata_expire=f"{metadata_expire}s" if metadata_expire else None,
        module_hotfixes=repo_cfg.get_module_hotfixes_option().get_value(),
        enabled=repo_cfg.get_enabled_option().get_value(),
        priority=repo_cfg.get_priority_option().get_value(),
        cost=repo_cfg.get_cost_option().get_value(),
        skip_if_unavailable=repo_cfg.get_skip_if_unavailable_option().get_value(),
        # username and password are intentionally omitted
    )


class DNF5(SolverBase):
    """Solver implements package related actions

    These include depsolving a package set, searching for packages, and dumping a list
    of all available packages.
    """

    SOLVER_NAME = "dnf5"

    _BASEARCH_MAP = _invert({
        'aarch64': ('aarch64',),
        'alpha': ('alpha', 'alphaev4', 'alphaev45', 'alphaev5', 'alphaev56',
                  'alphaev6', 'alphaev67', 'alphaev68', 'alphaev7', 'alphapca56'),
        'arm': ('armv5tejl', 'armv5tel', 'armv5tl', 'armv6l', 'armv7l', 'armv8l'),
        'armhfp': ('armv6hl', 'armv7hl', 'armv7hnl', 'armv8hl'),
        'i386': ('i386', 'athlon', 'geode', 'i386', 'i486', 'i586', 'i686'),
        'ia64': ('ia64',),
        'mips': ('mips',),
        'mipsel': ('mipsel',),
        'mips64': ('mips64',),
        'mips64el': ('mips64el',),
        'loongarch64': ('loongarch64',),
        'noarch': ('noarch',),
        'ppc': ('ppc',),
        'ppc64': ('ppc64', 'ppc64iseries', 'ppc64p7', 'ppc64pseries'),
        'ppc64le': ('ppc64le',),
        'riscv32': ('riscv32',),
        'riscv64': ('riscv64',),
        'riscv128': ('riscv128',),
        's390': ('s390',),
        's390x': ('s390x',),
        'sh3': ('sh3',),
        'sh4': ('sh4', 'sh4a'),
        'sparc': ('sparc', 'sparc64', 'sparc64v', 'sparcv8', 'sparcv9',
                  'sparcv9v'),
        'x86_64': ('x86_64', 'amd64', 'ia32e'),
    })

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        config: SolverConfig,
        persistdir: os.PathLike,
        license_index_path: Optional[os.PathLike] = None,
    ):
        super().__init__(config, persistdir, license_index_path)

        self.repos = self.config.repos or []
        self.request_repo_ids = {repo.repo_id for repo in self.repos}
        self.root_dir = self.config.root_dir

        self.base = dnf5.base.Base()

        # Base is the correct place to set substitutions, not per-repo.
        # See https://github.com/rpm-software-management/dnf5/issues/1248
        self.base.get_vars().set("arch", self.config.arch)
        self.base.get_vars().set("basearch", self._BASEARCH_MAP[self.config.arch])
        if self.config.releasever:
            self.base.get_vars().set('releasever', self.config.releasever)
        if self.config.proxy:
            self.base.get_vars().set('proxy', self.config.proxy)

        # Enable fastestmirror to ensure we choose the fastest mirrors for
        # downloading metadata (when depsolving) and downloading packages.
        conf = self.base.get_config()
        conf.fastestmirror = True

        # Weak dependencies are installed for the 1st transaction
        # This is set to False for any subsequent ones in depsolve()
        conf.install_weak_deps = True

        # We use the same cachedir for multiple architectures. Unfortunately,
        # this is something that doesn't work well in certain situations
        # with zchunk:
        # Imagine that we already have cache for arch1. Then, we use dnf-json
        # to depsolve for arch2. If ZChunk is enabled and available (that's
        # the case for Fedora), dnf will try to download only differences
        # between arch1 and arch2 metadata. But, as these are completely
        # different, dnf must basically redownload everything.
        # For downloding deltas, zchunk uses HTTP range requests. Unfortunately,
        # if the mirror doesn't support multi range requests, then zchunk will
        # download one small segment per a request. Because we need to update
        # the whole metadata (10s of MB), this can be extremely slow in some cases.
        # I think that we can come up with a better fix but let's just disable
        # zchunk for now. As we are already downloading a lot of data when
        # building images, I don't care if we download even more.
        conf.zchunk = False

        # Set the rest of the dnf configuration.
        if self.config.module_platform_id:
            conf.module_platform_id = self.config.module_platform_id
        conf.config_file_path = "/dev/null"
        conf.persistdir = persistdir
        conf.cachedir = self.config.cachedir

        # Include comps metadata by default
        metadata_types = ['comps']
        if self.config.optional_metadata:
            metadata_types.extend(self.config.optional_metadata)
        conf.optional_metadata_types = metadata_types

        try:
            for repo_conf in self.repos:
                self._dnfrepo(repo_conf)

            if self.root_dir:
                # This sets the varsdir to ("{root_dir}/usr/share/dnf5/vars.d/", "{root_dir}/etc/dnf/vars/") for custom
                # variable substitution (e.g. CentOS Stream 9's $stream variable)
                conf.installroot = self.root_dir
                conf.varsdir = (
                    os.path.join(self.root_dir, "etc/dnf/vars"),
                    os.path.join(self.root_dir, "usr/share/dnf5/vars.d")
                )

            # Cannot modify .conf() values after this
            # base.setup() should be called before loading repositories otherwise substitutions might not work.
            self.base.setup()

            if self.root_dir:
                repos_dir = os.path.join(self.root_dir, "etc/yum.repos.d")
                self.base.get_repo_sack().create_repos_from_dir(repos_dir)
                rq = dnf5.repo.RepoQuery(self.base)
                rq.filter_enabled(True)
                repo_iter = rq.begin()
                while repo_iter != rq.end():
                    repo = repo_iter.value()
                    repo_config = repo.get_config()
                    repo_config.sslcacert = modify_rootdir_path(
                        get_string_option(repo_config.get_sslcacert_option()),
                        self.root_dir,
                    )
                    repo_config.sslclientcert = modify_rootdir_path(
                        get_string_option(repo_config.get_sslclientcert_option()),
                        self.root_dir,
                    )
                    repo_config.sslclientkey = modify_rootdir_path(
                        get_string_option(repo_config.get_sslclientkey_option()),
                        self.root_dir,
                    )
                    repo_iter.next()

            self.base.get_repo_sack().load_repos(dnf5.repo.Repo.Type_AVAILABLE)
        except Exception as e:
            raise RepoError(e) from e

        if not any_repos_enabled(self.base):
            raise NoReposError("There are no enabled repositories")

    # pylint: disable=too-many-branches
    def _dnfrepo(self, desc: model.Repository, exclude_pkgs=None):
        """Makes a dnf.repo.Repo out of model.Repository configuration"""
        if not exclude_pkgs:
            exclude_pkgs = []

        sack = self.base.get_repo_sack()

        repo = sack.create_repo(desc.repo_id)
        conf = repo.get_config()

        # NB: default to using the repo_id as the name, to be consistent with DNF4 solver.
        # DNF5 documentation says that this should happen, but it doesn't seem to be the case.
        # https://github.com/rpm-software-management/dnf5/issues/2533
        conf.name = desc.repo_id
        if desc.name:
            conf.name = desc.name

        # NB: the fact that at least one is set is checked when parsing the Solver request
        if desc.baseurl:
            conf.baseurl = desc.baseurl
        if desc.metalink:
            conf.metalink = desc.metalink
        if desc.mirrorlist:
            conf.mirrorlist = desc.mirrorlist

        if desc.sslverify is not None:
            conf.sslverify = desc.sslverify
        if desc.sslcacert:
            conf.sslcacert = desc.sslcacert
        if desc.sslclientkey:
            conf.sslclientkey = desc.sslclientkey
        if desc.sslclientcert:
            conf.sslclientcert = desc.sslclientcert

        if desc.gpgcheck is not None:
            conf.gpgcheck = desc.gpgcheck
        if desc.repo_gpgcheck is not None:
            conf.repo_gpgcheck = desc.repo_gpgcheck
        if desc.gpgkey:
            # gpgkey can contain a full key, or it can be a URL
            # dnf expects urls, so write the key to a temporary location and add the file://
            # path to conf.gpgkey
            keydir = os.path.join(self.base.get_config().persistdir, "gpgkeys")
            if not os.path.exists(keydir):
                os.makedirs(keydir, mode=0o700, exist_ok=True)

            for key in desc.gpgkey:
                if key.startswith("-----BEGIN PGP PUBLIC KEY BLOCK-----"):
                    # Not using with because it needs to be a valid file for the duration. It
                    # is inside the temporary persistdir so will be cleaned up on exit.
                    # pylint: disable=consider-using-with
                    keyfile = tempfile.NamedTemporaryFile(dir=keydir, delete=False)
                    keyfile.write(key.encode("utf-8"))
                    conf.gpgkey += (f"file://{keyfile.name}",)
                    keyfile.close()
                else:
                    conf.gpgkey += (key,)

        if desc.metadata_expire:
            conf.metadata_expire = desc.metadata_expire
        if desc.module_hotfixes is not None:
            repo.module_hotfixes = desc.module_hotfixes
        if desc.enabled is not None:
            conf.enabled = desc.enabled
        if desc.priority is not None:
            conf.priority = desc.priority
        if desc.cost is not None:
            conf.cost = desc.cost
        if desc.username:
            conf.username = desc.username
        if desc.password:
            conf.password = desc.password
        if desc.skip_if_unavailable is not None:
            conf.skip_if_unavailable = desc.skip_if_unavailable

        # Set the packages to exclude
        conf.excludepkgs = exclude_pkgs

        return repo

    def _sbom_for_pkgset(self, pkgset: List[dnf5.rpm.Package]) -> Dict:
        """
        Create an SBOM document for the given package set.

        For now, only SPDX v2 is supported.
        """
        pkgset = dnf_pkgset_to_sbom_pkgset(pkgset)
        spdx_doc = sbom_pkgset_to_spdx2_doc(pkgset, self.license_index_path)
        return spdx_doc.to_dict()

    def _dnf_repo_to_repository(self, repo: dnf5.repo.Repo) -> model.Repository:
        """
        Convert a dnf5.repo.Repo to a Repository.
        """
        repo_model = _dnf_repo_to_repository(repo, self.root_dir, self.request_repo_ids)
        self.set_rhsm_flag(repo_model)
        return repo_model

    def dump(self) -> model.DumpResult:
        """dump returns a list of all available packages"""
        seen_repos = {}
        repositories = []

        def package_iter():
            q = dnf5.rpm.PackageQuery(self.base)
            q.filter_available()
            for package in q:
                repo_id = package.get_repo().get_id()
                if repo_id not in seen_repos:
                    repo = self._dnf_repo_to_repository(package.get_repo())
                    seen_repos[repo_id] = repo
                    repositories.append(repo)
                yield _dnf_pkg_to_package(package)

        return model.DumpResult(package_iter(), repositories)

    def search(self, args: SearchCmdArgs) -> model.SearchResult:
        """ Perform a search on the available packages"""

        packages = []
        repositories = {}

        # NOTE: Build query one piece at a time, don't pass all to filterm at the same
        # time.
        for name in args.packages:
            q = dnf5.rpm.PackageQuery(self.base)
            q.filter_available()

            # If the package name glob has * in it, use glob.
            # If it has *name* use substr
            # If it has neither use exact match
            if "*" in name:
                if name[0] != "*" or name[-1] != "*":
                    q.filter_name([name], GLOB)
                else:
                    q.filter_name([name.replace("*", "")], CONTAINS)
            else:
                q.filter_name([name], EQ)

            if args.latest:
                q.filter_latest_evr()

            for package in list(q):
                packages.append(_dnf_pkg_to_package(package))
                if package.get_repo().get_id() not in repositories:
                    repositories[package.get_repo().get_id()] = self._dnf_repo_to_repository(package.get_repo())

        return model.SearchResult(packages, list(repositories.values()))

    def depsolve(self, args: DepsolveCmdArgs) -> model.DepsolveResult:
        """Perform a dependency resolution for the given transactions"""
        last_dnf_transaction: List[dnf5.rpm.Package] = []
        # List of transaction results, each containing a list of packages that are a result of dependency resolution.
        # Each transaction result is a superset of the previous transaction result.
        # The package list in each transaction is alphabetically sorted by full NEVRA.
        transactions_results: List[List[model.Package]] = []
        repositories_by_id: Dict[str, model.Repository] = {}
        repositories: List[model.Repository] = []

        for transaction in args.transactions:
            goal = dnf5.base.Goal(self.base)
            goal.reset()
            sack = self.base.get_rpm_package_sack()
            sack.clear_user_excludes()

            # Restrict dependency resolution to only use packages from the
            # repos listed in repo_ids by excluding all packages from other
            # repos from the sack. Without this, dependencies would be
            # resolved from all enabled repos regardless of repo_ids.
            if transaction.repo_ids:
                allowed_repo_ids = set(transaction.repo_ids)
                rq = dnf5.repo.RepoQuery(self.base)
                rq.filter_enabled(True)
                # libdnf5's RepoQuery doesn't support Python iteration
                repo_iter = rq.begin()
                while repo_iter != rq.end():
                    repo = repo_iter.value()
                    if repo.get_id() not in allowed_repo_ids:
                        q = dnf5.rpm.PackageQuery(self.base)
                        q.filter_available()
                        q.filter_repo_id([repo.get_id()], EQ)
                        sack.add_user_excludes(q)
                    repo_iter.next()

            # weak deps are selected per-transaction
            self.base.get_config().install_weak_deps = transaction.install_weak_deps

            # set the packages from the last transaction as installed
            for installed_pkg in last_dnf_transaction:
                goal.add_rpm_install(installed_pkg)

            # NOTE: DNF4's install_specs() accepts exclude_specs directly and handles exclusion internally using
            # sack.add_excludes(). DNF5's Goal API has no equivalent -- add_install() only accepts package specs,
            # not excludes. To achieve the same effect, we explicitly exclude matching packages from the sack using
            # sack.add_user_excludes(), the same mechanism used for repo_ids filtering above. The excludes are scoped
            # per-transaction because sack.clear_user_excludes() is called at the start of each iteration. This block
            # is placed after add_rpm_install() (which re-adds packages from previous transactions) to match DNF4's
            # ordering, where package_install() precedes install_specs(). Note that sack-level excludes do not affect
            # packages already added to the goal via add_rpm_install(), so excludes only influence the resolution
            # of newly requested packages.
            if transaction.exclude_specs:
                for exclude_spec in transaction.exclude_specs:
                    q = dnf5.rpm.PackageQuery(self.base)
                    q.filter_available()
                    q.filter_name([exclude_spec], GLOB)
                    sack.add_user_excludes(q)

            # Support group/environment names as well as ids
            settings = dnf5.base.GoalJobSettings()
            settings.group_with_name = True

            for package_spec in transaction.package_specs:
                goal.add_install(package_spec, settings)
            goal_result = goal.resolve()

            transaction_problems = goal_result.get_problems()
            if transaction_problems != NO_PROBLEM:
                error_msg = "\n".join(goal_result.get_resolve_logs_as_strings())
                if not (transaction_problems & ~MARKING_PROBLEMS):
                    raise MarkingError(error_msg)
                raise DepsolveError(error_msg)

            transaction_result = []
            # store the current transaction result
            last_dnf_transaction.clear()
            for tsi in goal_result.get_transaction_packages():
                # Only add packages being installed, upgraded, downgraded, or reinstalled
                if not dnf5.base.transaction.transaction_item_action_is_inbound(tsi.get_action()):
                    continue
                pkg = tsi.get_package()
                last_dnf_transaction.append(pkg)
                transaction_result.append(_dnf_pkg_to_package(pkg))
                if pkg.get_repo().get_id() not in repositories_by_id:
                    repositories_by_id[pkg.get_repo().get_id()] = self._dnf_repo_to_repository(pkg.get_repo())

            # NB: DNF5 solver returns packages in topological order, but the original DNF4 solver returns
            # alphabetically sorted packages. To be consistent and match the original DNF4 solver behavior,
            # we sort the packages alphabetically by full NEVRA.
            # NB: the org.osbuild.rpm stage as generated by osbuild/images does not depend on the order of packages,
            # because rpm gets the full package set at once and it will reorder the packages as needed when installing.
            transaction_result.sort()
            transactions_results.append(transaction_result)

        # NB: we sort the repositories by repo_id to ensure consistent ordering across DNF4 and DNF5.
        repositories = list(repositories_by_id.values())
        repositories.sort(key=lambda x: x.repo_id)

        # Something went wrong, but no error was generated by goal.resolve()
        if len(args.transactions) > 0 and len(transactions_results[-1]) == 0:
            raise DepsolveError("Empty transaction results")

        sbom = None
        if args.sbom_request:
            sbom = self._sbom_for_pkgset(last_dnf_transaction)

        return model.DepsolveResult(
            transactions=transactions_results,
            repositories=repositories,
            modules=None,  # DNF5 Solver does not support modules
            sbom=sbom,
        )



#!/usr/bin/python3

import pytest

from osbuild import testutil

STAGE_NAME = "org.osbuild.ovf"


# Prepare dataset containing good and bad API call parameters
@pytest.mark.parametrize("test_data, expected_err", [
    # Bad API parameters
    ({}, "'vmdk' is a required property"),
    ({"vmdk": 123}, "123 is not of type 'string'"),
    ({"vmdk": "imagename"}, "'imagename' does not match '[a-zA-Z0-9+_.-]+.vmdk'"),
    ({"vmdk": "imagename.vmdk", "vmware": {"os_type": 1}}, "is not of type 'string'"),
    ({"vmdk": "imagename.vmdk", "virtualbox": {"os_type": 1}}, "is not of type 'string'"),
    ({"vmdk": "imagename.vmdk", "virtualbox": {"mac_address": 1}}, "is not of type 'string'"),
    ({"vmdk": "imagename.vmdk", "virtualbox": {"mac_address": "1"}}, "'1' does not match '^[a-fA-F0-9]{12}$'"),
    ({"vmdk": "imagename.vmdk", "vmware": {"virtual_hardware_version": 1}}, "is not of type 'string'"),
    ({"vmdk": "imagename.vmdk", "vmware": {"virtual_hardware_version": "21"}}, "'21' does not match '^vmx-[0-9]+$'"),
    # Good API parameters
    ({"vmdk": "imagename.vmdk"}, ""),
    ({"vmdk": "imagename.vmdk", "vmware": {"os_type": "OtherGuest",
     "virtual_hardware_version": "vmx-21"}, "virtualbox": {"os_type": "OtherGuest"}}, ""),
])
# This test validates only API calls using correct and incorrect queries
def test_schema_validation_ovf(stage_schema, test_data, expected_err):
    test_input = {
        "type": STAGE_NAME,
        "devices": {
            "device": {
                "path": "some-path",
            },
        },
        "options": {
        }
    }
    test_input["options"].update(test_data)
    res = stage_schema.validate(test_input)

    if expected_err == "":
        assert res.valid is True, f"err: {[e.as_dict() for e in res.errors]}"
    else:
        assert res.valid is False
        testutil.assert_jsonschema_error_contains(res, expected_err, expected_num_errs=1)


@pytest.mark.skipif(not testutil.has_executable("qemu-img"), reason="need qemu-img")
@pytest.mark.parametrize("test_opts, expected_substrings", [
    # Replacements
    ({"vmware": {"os_type": "my-vmware-os-type"}}, ["my-vmware-os-type"],),
    ({"virtualbox": {"os_type": "my-vbox-os-type"}}, ["my-vbox-os-type"],),
    ({"virtualbox": {"mac_address": "000000000000"}}, ["000000000000"],),
])
def test_ovf_default_template(tmp_path, stage_module, test_opts, expected_substrings):
    faked_vmdk_path = tmp_path / "some-image.vmdk"
    faked_vmdk_path.write_bytes(b"1234")

    opts = {
        "vmdk": faked_vmdk_path,
    }
    opts.update(test_opts)

    stage_module.main(opts, tmp_path)

    expected_template_path = tmp_path / "some-image.ovf"

    for substring in expected_substrings:
        assert substring in expected_template_path.read_text()


def test_ovf_vbox_mac_address_differs(stage_module):
    macs = [
        stage_module.vbox_mac_address({}),
        stage_module.vbox_mac_address({}),
    ]

    assert macs[0] != macs[1]
    assert all(m.startswith("080027") for m in macs)
    assert all(len(m) == 12 for m in macs)


def test_ovf_vbox_mac_address_options(stage_module):
    assert stage_module.vbox_mac_address({"virtualbox": {"mac_address": "000000000000"}}) == "000000000000"



""" Version 2 of the manifest description

Second, and current, version of the manifest description
"""
from typing import Any, Dict, Optional

from osbuild.meta import Index, ModuleInfo, ValidationResult

from ..inputs import Input
from ..objectstore import ObjectStore
from ..pipeline import Manifest, ManifestBuildResult, Pipeline, Runner, Stage
from ..sources import Source

VERSION = "2"


# pylint: disable=too-many-statements
def describe(manifest: Manifest, *, with_id=False) -> Dict:

    # Undo the build, runner pairing introduce by the loading
    # code. See the comment there for more details
    runners = {
        p.build: p.runner for p in manifest.pipelines.values()
        if p.build
    }

    def pipeline_ref(pid):
        if with_id:
            return pid

        pl = manifest[pid]
        return f"name:{pl.name}"

    def describe_device(dev):
        desc = {
            "type": dev.info_name
        }

        if dev.options:
            desc["options"] = dev.options

        return desc

    def describe_devices(devs: Dict):
        desc = {
            name: describe_device(dev)
            for name, dev in devs.items()
        }
        return desc

    def describe_input(ip: Input):
        origin = ip.origin
        desc = {
            "type": ip.info_name,
            "origin": origin,
        }
        if ip.options:
            desc["options"] = ip.options

        refs = {}
        for name, ref in ip.refs.items():
            if origin == "org.osbuild.pipeline":
                name = pipeline_ref(name)
            refs[name] = ref

        if refs:
            desc["references"] = refs

        return desc

    def describe_inputs(ips: Dict[str, Input]):
        desc = {
            name: describe_input(ip)
            for name, ip in ips.items()
        }
        return desc

    def describe_mount(mnt):
        desc = {
            "name": mnt.name,
            "type": mnt.info_name,
            "target": mnt.target
        }

        if mnt.device:
            desc["source"] = mnt.device.name
        if mnt.options:
            desc["options"] = mnt.options
        if mnt.partition:
            desc["partition"] = mnt.partition
        return desc

    def describe_mounts(mounts: Dict):
        desc = [
            describe_mount(mnt)
            for mnt in mounts.values()
        ]
        return desc

    def describe_stage(s: Stage):
        desc = {
            "type": s.info_name
        }

        if with_id:
            desc["id"] = s.id

        if s.options:
            desc["options"] = s.options

        devs = describe_devices(s.devices)
        if devs:
            desc["devices"] = devs

        mounts = describe_mounts(s.mounts)
        if mounts:
            desc["mounts"] = mounts

        ips = describe_inputs(s.inputs)
        if ips:
            desc["inputs"] = ips

        return desc

    def describe_pipeline(p: Pipeline):
        desc: Dict[str, Any] = {
            "name": p.name
        }

        if p.build:
            desc["build"] = pipeline_ref(p.build)

        runner = runners.get(p.id)
        if runner:
            desc["runner"] = runner.name

        stages = [
            describe_stage(stage)
            for stage in p.stages
        ]

        if stages:
            desc["stages"] = stages

        return desc

    def describe_source(s: Source):
        desc = {
            "items": s.items
        }

        return desc

    pipelines = [
        describe_pipeline(pipeline)
        for pipeline in manifest.pipelines.values()
    ]

    sources = {
        source.info_name: describe_source(source)
        for source in manifest.sources
    }

    description: Dict[str, Any] = {
        "version": VERSION,
        "pipelines": pipelines
    }

    if manifest.metadata:
        description["metadata"] = manifest.metadata

    if sources:
        description["sources"] = sources

    return description


def resolve_ref(name: str, manifest: Manifest) -> str:
    ref = name[5:]
    target = manifest.pipelines.get(ref)
    if not target:
        raise ValueError(f"Unknown pipeline reference: name:{ref}")
    return target.id


def sort_devices(devices: Dict) -> Dict:
    """Sort the devices so that dependencies are in the correct order

    We need to ensure that parents are sorted before the devices that
    depend on them. For this we keep a list of devices that need to
    be processed and iterate over that list as long as it has devices
    in them and we make progress, i.e. the length changes.
    """
    result = {}
    todo = list(devices.keys())

    while todo:
        before = len(todo)

        for i, name in enumerate(todo):
            desc = devices[name]

            parent = desc.get("parent")
            if parent and parent not in result:
                # if the parent is not in the `result` list, it must
                # be in `todo`; otherwise it is missing
                if parent not in todo:
                    msg = f"Missing parent device '{parent}' for '{name}'"
                    raise ValueError(msg)

                continue

            # no parent, or parent already present, ok to add to the
            # result and "remove" from the todo list, by setting the
            # contents to `None`.
            result[name] = desc
            todo[i] = None

        todo = list(filter(bool, todo))
        if len(todo) == before:
            # we made no progress, which means that all devices in todo
            # depend on other devices in todo, hence we have a cycle
            raise ValueError("Cycle detected in 'devices'")

    return result


def load_device(name: str, description: Dict, index: Index, stage: Stage):
    device_type = description["type"]
    options = description.get("options", {})
    parent = description.get("parent")

    if parent:
        device = stage.devices.get(parent)
        if not parent:
            raise ValueError(f"Unknown parent device: {parent}")
        parent = device

    info = index.get_module_info("Device", device_type)

    if not info:
        raise TypeError(f"Missing meta information for {device_type}")
    stage.add_device(name, info, parent, options)


def load_input(name: str, description: Dict, index: Index, stage: Stage, manifest: Manifest, source_refs: set):
    input_type = description["type"]
    origin = description["origin"]
    options = description.get("options", {})

    info = index.get_module_info("Input", input_type)
    ip = stage.add_input(name, info, origin, options)

    refs = description.get("references", {})

    if isinstance(refs, list):
        def make_ref(ref):
            if isinstance(ref, str):
                return ref, {}
            if isinstance(ref, dict):
                return ref.get("id"), ref.get("options", {})
            raise ValueError(f"Invalid reference: {ref}")

        refs = dict(make_ref(ref) for ref in refs)

    if origin == "org.osbuild.pipeline":
        resolved = {}
        for r, desc in refs.items():
            if not r.startswith("name:"):
                continue
            target = resolve_ref(r, manifest)
            resolved[target] = desc
        refs = resolved
    elif origin == "org.osbuild.source":
        unknown_refs = set(refs.keys()) - source_refs
        if unknown_refs:
            raise ValueError(f"Unknown source reference(s) {unknown_refs}")

    for r, desc in refs.items():
        ip.add_reference(r, desc)


def load_mount(description: Dict, index: Index, stage: Stage):
    mount_type = description["type"]
    info = index.get_module_info("Mount", mount_type)

    name = description["name"]

    if name in stage.mounts:
        raise ValueError(f"Duplicated mount '{name}'")

    source = description.get("source")
    partition = description.get("partition")
    target = description.get("target")

    options = description.get("options", {})

    device = None
    if source:
        device = stage.devices.get(source)
        if not device:
            raise ValueError(f"Unknown device '{source}' for mount '{name}'")

    stage.add_mount(name, info, device, partition, target, options)


def load_stage(description: Dict, index: Index, pipeline: Pipeline, manifest: Manifest, source_refs):
    stage_type = description["type"]
    opts = description.get("options", {})
    info = index.get_module_info("Stage", stage_type)

    stage = pipeline.add_stage(info, opts)

    devs = description.get("devices", {})
    devs = sort_devices(devs)

    for name, desc in devs.items():
        load_device(name, desc, index, stage)

    ips = description.get("inputs", {})
    for name, desc in ips.items():
        load_input(name, desc, index, stage, manifest, source_refs)

    mounts = description.get("mounts", [])
    for mount in mounts:
        load_mount(mount, index, stage)

    return stage


def load_pipeline(description: Dict, index: Index, manifest: Manifest, source_refs: set):
    name = description["name"]
    build = description.get("build")
    source_epoch = description.get("source-epoch")

    if build and build.startswith("name:"):
        target = resolve_ref(build, manifest)
        build = target

    # NB: The runner mapping will later be changed in `load`.
    # The host runner here is just to always have a Runner
    # (instead of a Optional[Runner]) to make mypy happy
    runner_name = description.get("runner")
    runner = None
    if runner_name:
        runner = Runner(index.detect_runner(runner_name), runner_name)
    else:
        runner = Runner(index.detect_host_runner())

    pl = manifest.add_pipeline(name, runner, build, source_epoch)

    for desc in description.get("stages", []):
        load_stage(desc, index, pl, manifest, source_refs)


def load(description: Dict, index: Index) -> Manifest:
    """Load a manifest description"""

    sources = description.get("sources", {})
    pipelines = description.get("pipelines", [])
    metadata = description.get("metadata", {})

    manifest = Manifest()
    source_refs = set()

    # metadata
    for key, value in metadata.items():
        manifest.add_metadata(key, value)

    # load the sources
    for name, desc in sources.items():
        info = index.get_module_info("Source", name)
        items = desc.get("items", {})
        options = desc.get("options", {})
        manifest.add_source(info, items, options)
        source_refs.update(items.keys())

    for desc in pipelines:
        load_pipeline(desc, index, manifest, source_refs)

    # The "runner" property in the manifest format is the
    # runner to the run the pipeline with. In osbuild the
    # "runner" property belongs to the "build" pipeline,
    # i.e. is what runner to use for it. This we have to
    # go through the pipelines and fix things up
    pipelines = manifest.pipelines.values()

    host_runner = Runner(index.detect_host_runner())
    runners = {
        pl.id: pl.runner for pl in pipelines
    }

    for pipeline in pipelines:
        if not pipeline.build:
            pipeline.runner = host_runner
            continue

        runner = runners[pipeline.build]
        pipeline.runner = runner

    return manifest


# pylint: disable=too-many-branches
def output(manifest: Manifest, res: ManifestBuildResult, store: Optional[ObjectStore] = None) -> Dict:
    """Convert a result into the v2 format"""

    def collect_metadata(p: Pipeline) -> Dict[str, Any]:
        data: Dict[str, Any] = {}

        if not store:  # for testing
            return data

        obj = store.get(p.id)
        if not obj:
            return data

        for stage in p.stages:
            md = obj.meta.get(stage.id)
            if not md:
                continue
            val = data.setdefault(stage.name, {})
            val.update(md)

        return data

    result: Dict[str, Any] = {}

    if not res.success:
        if not res.download_result.success:
            failing_sources = list(filter(lambda src: not src[1].success, res.download_result.sources.items()))
            if not failing_sources:
                raise RuntimeError("sources marked as failed but no failing sources found in download result")
            error = ''
            failing_source_items = list(filter(lambda src: not src.success, failing_sources[0][1].results))
            if not failing_source_items:
                error = "cannot find failing source item"
            else:
                error = failing_source_items[0].error
            result = {
                "type": "error",
                "success": False,
                "error": {
                    "type": "org.osbuild.error.source",
                    "details": {
                        "source": {
                            "type": failing_sources[0][0],
                            "output": error,
                        },
                    },
                },
            }
        else:
            pr = res.pipeline_results
            last = list(pr.keys())[-1]
            failed = pr[last].stages[-1]

            result = {
                "type": "error",
                "success": False,
                "error": {
                    "type": "org.osbuild.error.stage",
                    "details": {
                        "stage": {
                            "id": failed.id,
                            "type": failed.name,
                            "output": failed.output,
                            "error": failed.error,
                        }
                    }
                }
            }
    else:
        result = {
            "type": "result",
            "success": True,
            "metadata": {}
        }

        # gather all the metadata
        for p in manifest.pipelines.values():
            data: Dict[str, Any] = collect_metadata(p)
            if data:
                result["metadata"][p.name] = data

    # generate the sources
    result["sources"] = {
        "success": res.download_result.success,
    }
    for src_type, src in res.download_result.sources.items():
        result["sources"][src_type] = []
        for src_item_res in src.results:
            src_output = {}
            if "cached" in src_item_res.metadata:
                src_output["cached"] = src_item_res.metadata["cached"]
            if "checksum" in src_item_res.metadata:
                src_output["checksum"] = src_item_res.metadata["checksum"]
            if src_item_res.error:
                src_output["error"] = src_item_res.error
            result["sources"][src_type].append(src_output)

    # generate the log
    result["log"] = {}
    for p in manifest.pipelines.values():
        if p.id not in res.pipeline_results:
            continue

        r = res.pipeline_results[p.id]
        log = []

        for stage in r.stages:
            data = {
                "id": stage.id,
                "type": stage.name,
                "output": stage.output,
            }
            if not stage.success:
                data["success"] = stage.success
                if stage.error:
                    data["error"] = stage.error

            log.append(data)

        if log:
            result["log"][p.name] = log

    return result


def validate(manifest: Dict, index: Index) -> ValidationResult:

    schema = index.get_schema("Manifest", version="2")
    result = schema.validate(manifest)

    def validate_module(mod, klass, path):
        name = mod.get("type")
        if not name:
            return
        schema = index.get_schema(klass, name, version="2")
        res = schema.validate(mod)
        result.merge(res, path=path)

    def validate_stage_modules(klass, stage, path):
        group = ModuleInfo.MODULES[klass]
        items = stage.get(group, {})

        if isinstance(items, list):
            items = {i["name"]: i for i in items}

        for name, mod in items.items():
            validate_module(mod, klass, path + [group, name])

    def validate_stage(stage, path):
        name = stage["type"]
        schema = index.get_schema("Stage", name, version="2")
        res = schema.validate(stage)
        result.merge(res, path=path)

        for mod in ("Device", "Input", "Mount"):
            validate_stage_modules(mod, stage, path)

    def validate_pipeline(pipeline, path):
        stages = pipeline.get("stages", [])
        for i, stage in enumerate(stages):
            validate_stage(stage, path + ["stages", i])

    # sources
    sources = manifest.get("sources", {})
    for name, source in sources.items():
        schema = index.get_schema("Source", name, version="2")
        res = schema.validate(source)
        result.merge(res, path=["sources", name])

    # pipelines
    pipelines = manifest.get("pipelines", [])
    for i, pipeline in enumerate(pipelines):
        validate_pipeline(pipeline, path=["pipelines", i])

    return result



#!/usr/bin/python3

import os.path

import pytest

from osbuild.testutil import make_fake_tree

STAGE_NAME = "org.osbuild.grub2"


def test_grub2_copy_efi_data(tmp_path, stage_module):
    fake_efi_src_dir = tmp_path / "fake-efi/EFI"
    make_fake_tree(fake_efi_src_dir, {
        "fedora/a.shim": "fake shim",
        "BOOT/BOOTX64.EFI": "I'm not real",
    })

    test_options = {
        "rootfs": {
            "label": "my-rootfs",
        },
        "uefi": {
            "install": True,
            "efi_src_dir": fake_efi_src_dir,
            "vendor": "fedora",
        },
    }
    fake_tree = tmp_path / "tree"
    stage_module.main(fake_tree, test_options)
    assert (fake_tree / "boot/efi/EFI/fedora/a.shim").exists()
    assert (fake_tree / "boot/efi/EFI/BOOT/BOOTX64.EFI").exists()


# Test that the /etc/default/grub file is created with the correct content
@pytest.mark.parametrize("test_data,kernel_opts,expected_conf,expected_boot_conf", [
    # default
    ({}, "", """GRUB_CMDLINE_LINUX=""
GRUB_TIMEOUT=0
GRUB_ENABLE_BLSCFG=true
""",
        """
set timeout=0

# load the grubenv file
load_env

# selection of the next boot entry via variables 'next_entry' and
# `saved_entry` present in the 'grubenv' file. Both variables are
# set by grub tools, like grub2-reboot, grub2-set-default

if [ "${next_entry}" ] ; then
   set default="${next_entry}"
   set next_entry=
   save_env next_entry
   set boot_once=true
else
   set default="${saved_entry}"
fi

search --no-floppy --set=root --label root
set boot=${root}
function load_video {
  insmod all_video
}
${serial}${terminal_input}${terminal_output}
blscfg
"""),
    # custom
    ({
        "default": "0",
        "timeout": 10,
        "terminal_input": ["console"],
        "terminal_output": ["serial"],
        "serial": "serial --speed=115200 --unit=0 --word=8 --parity=no --stop=1",
    }, "", """GRUB_CMDLINE_LINUX=""
GRUB_TIMEOUT=10
GRUB_ENABLE_BLSCFG=true
GRUB_SERIAL_COMMAND="serial --speed=115200 --unit=0 --word=8 --parity=no --stop=1"
GRUB_TERMINAL_INPUT="console"
GRUB_TERMINAL_OUTPUT="serial"
GRUB_DEFAULT=0
""",
        """
set timeout=10

# load the grubenv file
load_env

# selection of the next boot entry via variables 'next_entry' and
# `saved_entry` present in the 'grubenv' file. Both variables are
# set by grub tools, like grub2-reboot, grub2-set-default

if [ "${next_entry}" ] ; then
   set default="${next_entry}"
   set next_entry=
   save_env next_entry
   set boot_once=true
else
   set default="${saved_entry}"
fi

search --no-floppy --set=root --label root
set boot=${root}
function load_video {
  insmod all_video
}

serial --speed=115200 --unit=0 --word=8 --parity=no --stop=1
terminal_input console
terminal_output serial
blscfg
"""),
    # custom (Azure)
    ({
        "default": "saved",
        "disable_submenu": True,
        "disable_recovery": True,
        "distributor": "$(sed 's, release .*$,,g' /etc/system-release)",
        "serial": "serial --speed=115200 --unit=0 --word=8 --parity=no --stop=1",
        "terminal": ["serial", "console"],
        "terminal_output": ["console"],
        "timeout": 10,
        "timeout_style": "countdown",
    },
        "loglevel=3 crashkernel=auto console=tty1 console=ttyS0 earlyprintk=ttyS0 rootdelay=300",
        """GRUB_CMDLINE_LINUX="loglevel=3 crashkernel=auto console=tty1 console=ttyS0 earlyprintk=ttyS0 rootdelay=300"
GRUB_TIMEOUT=10
GRUB_ENABLE_BLSCFG=true
GRUB_DISABLE_RECOVERY=true
GRUB_DISABLE_SUBMENU=true
GRUB_DISTRIBUTOR="$(sed 's, release .*$,,g' /etc/system-release)"
GRUB_SERIAL_COMMAND="serial --speed=115200 --unit=0 --word=8 --parity=no --stop=1"
GRUB_TERMINAL="serial console"
GRUB_TERMINAL_OUTPUT="console"
GRUB_TIMEOUT_STYLE=countdown
GRUB_DEFAULT=saved
""",
        """
set timeout=10

# load the grubenv file
load_env

# selection of the next boot entry via variables 'next_entry' and
# `saved_entry` present in the 'grubenv' file. Both variables are
# set by grub tools, like grub2-reboot, grub2-set-default

if [ "${next_entry}" ] ; then
   set default="${next_entry}"
   set next_entry=
   save_env next_entry
   set boot_once=true
else
   set default="${saved_entry}"
fi

search --no-floppy --set=root --label root
set boot=${root}
function load_video {
  insmod all_video
}

serial --speed=115200 --unit=0 --word=8 --parity=no --stop=1${terminal_input}
terminal_output console
blscfg
"""),
])
def test_grub2_default_conf(tmp_path, stage_module, test_data, kernel_opts, expected_conf, expected_boot_conf):
    treedir = tmp_path / "tree"
    confpath = treedir / "etc/default/grub"
    confpath.parent.mkdir(parents=True, exist_ok=True)
    bootconfpath = treedir / "boot/grub2/grub.cfg"

    options = {
        "rootfs": {
            "label": "root"
        },
        "entries": [
            {
                "id": "fff",
                "kernel": "4.18",
                "product": {
                    "name": "Fedora",
                    "version": "40"
                }
            }
        ]
    }

    options["config"] = test_data
    options["kernel_opts"] = kernel_opts
    stage_module.main(treedir, options)

    assert os.path.exists(confpath)
    assert confpath.read_text() == expected_conf

    assert os.path.exists(bootconfpath)
    assert bootconfpath.read_text() == expected_boot_conf


# Test that the /etc/default/grub file is created with the correct content
@pytest.mark.parametrize("test_data,kernel_opts,expected_conf,expected_boot_conf", [
    # default
    ({}, "", """GRUB_CMDLINE_LINUX=""
GRUB_TIMEOUT=0
GRUB_ENABLE_BLSCFG=true
""",
        """
set timeout=0

# load the grubenv file
load_env

# selection of the next boot entry via variables 'next_entry' and
# `saved_entry` present in the 'grubenv' file. Both variables are
# set by grub tools, like grub2-reboot, grub2-set-default

if [ "${next_entry}" ] ; then
   set default="${next_entry}"
   set next_entry=
   save_env next_entry
   set boot_once=true
else
   set default="${saved_entry}"
fi

search --no-floppy --set=root --label root
set boot=${root}
function load_video {
  insmod all_video
}

blscfg
"""),
    # custom
    ({
        "default": "0",
        "timeout": 10,
        "terminal_input": ["console"],
        "terminal_output": ["serial"],
        "serial": "serial --speed=115200 --unit=0 --word=8 --parity=no --stop=1",
    }, "", """GRUB_CMDLINE_LINUX=""
GRUB_TIMEOUT=10
GRUB_ENABLE_BLSCFG=true
GRUB_SERIAL_COMMAND="serial --speed=115200 --unit=0 --word=8 --parity=no --stop=1"
GRUB_TERMINAL_INPUT="console"
GRUB_TERMINAL_OUTPUT="serial"
GRUB_DEFAULT=0
""",
        """
set timeout=10

# load the grubenv file
load_env

# selection of the next boot entry via variables 'next_entry' and
# `saved_entry` present in the 'grubenv' file. Both variables are
# set by grub tools, like grub2-reboot, grub2-set-default

if [ "${next_entry}" ] ; then
   set default="${next_entry}"
   set next_entry=
   save_env next_entry
   set boot_once=true
else
   set default="${saved_entry}"
fi

search --no-floppy --set=root --label root
set boot=${root}
function load_video {
  insmod all_video
}

serial --speed=115200 --unit=0 --word=8 --parity=no --stop=1
terminal_input console
terminal_output serial
blscfg
"""),
    # custom (Azure)
    ({
        "default": "saved",
        "disable_submenu": True,
        "disable_recovery": True,
        "distributor": "$(sed 's, release .*$,,g' /etc/system-release)",
        "serial": "serial --speed=115200 --unit=0 --word=8 --parity=no --stop=1",
        "terminal": ["serial", "console"],
        "terminal_output": ["console"],
        "timeout": 10,
        "timeout_style": "countdown",
    },
        "loglevel=3 crashkernel=auto console=tty1 console=ttyS0 earlyprintk=ttyS0 rootdelay=300",
        """GRUB_CMDLINE_LINUX="loglevel=3 crashkernel=auto console=tty1 console=ttyS0 earlyprintk=ttyS0 rootdelay=300"
GRUB_TIMEOUT=10
GRUB_ENABLE_BLSCFG=true
GRUB_DISABLE_RECOVERY=true
GRUB_DISABLE_SUBMENU=true
GRUB_DISTRIBUTOR="$(sed 's, release .*$,,g' /etc/system-release)"
GRUB_SERIAL_COMMAND="serial --speed=115200 --unit=0 --word=8 --parity=no --stop=1"
GRUB_TERMINAL="serial console"
GRUB_TERMINAL_OUTPUT="console"
GRUB_TIMEOUT_STYLE=countdown
GRUB_DEFAULT=saved
""",
        """
set timeout=10

# load the grubenv file
load_env

# selection of the next boot entry via variables 'next_entry' and
# `saved_entry` present in the 'grubenv' file. Both variables are
# set by grub tools, like grub2-reboot, grub2-set-default

if [ "${next_entry}" ] ; then
   set default="${next_entry}"
   set next_entry=
   save_env next_entry
   set boot_once=true
else
   set default="${saved_entry}"
fi

search --no-floppy --set=root --label root
set boot=${root}
function load_video {
  insmod all_video
}

serial --speed=115200 --unit=0 --word=8 --parity=no --stop=1
terminal_output console
blscfg
"""),
])
def test_grub2_default_conf_v2(tmp_path, stage_module, test_data, kernel_opts, expected_conf, expected_boot_conf):
    treedir = tmp_path / "tree"
    confpath = treedir / "etc/default/grub"
    confpath.parent.mkdir(parents=True, exist_ok=True)
    bootconfpath = treedir / "boot/grub2/grub.cfg"

    options = {
        "compat_version": 2,
        "rootfs": {
            "label": "root"
        },
        "entries": [
            {
                "id": "fff",
                "kernel": "4.18",
                "product": {
                    "name": "Fedora",
                    "version": "40"
                }
            }
        ],
    }

    options["config"] = test_data
    options["kernel_opts"] = kernel_opts
    stage_module.main(treedir, options)

    assert os.path.exists(confpath)
    assert confpath.read_text() == expected_conf

    assert os.path.exists(bootconfpath)
    assert bootconfpath.read_text() == expected_boot_conf



#!/usr/bin/python3

import os.path
import re
import subprocess
import tarfile

import pytest

from osbuild import testutil
from osbuild.testutil import has_executable, make_fake_input_tree

STAGE_NAME = "org.osbuild.tar"


@pytest.mark.parametrize("test_data,expected_err", [
    # bad
    ({}, "'filename' is a required property"),
    ({
        "filename": "out.tar",
        "root-node": "include",
        "paths": ["file1"]
    }, "{'filename': 'out.tar', 'root-node': 'include', 'paths': ['file1']} is not valid under any of the given schemas"),
    ({
        "filename": "foo",
        "transform": ["transform-cannot-be-passed-multiple-times"],
    }, " is not of type 'string'"),
    ({
        "filename": "foo",
        "compression": "EXTREME",
    }, "'EXTREME' is not one of"),
    # good
    ({"filename": "out.tar", "root-node": "include"}, ""),
    ({"filename": "out.tar", "paths": ["file1"]}, ""),
    ({"filename": "out.tar", "sparse": True}, ""),
    ({"filename": "out.tar"}, ""),
    ({"filename": "out.tar", "transform": "s/foo/bar"}, ""),
    ({"filename": "out.tar", "compression": "auto"}, ""),
    ({"filename": "out.tar", "compression": "gzip"}, ""),
])
def test_schema_validation_tar(stage_schema, test_data, expected_err):
    test_input = {
        "type": STAGE_NAME,
        "options": {
        }
    }
    test_input["options"].update(test_data)
    res = stage_schema.validate(test_input)

    if expected_err == "":
        assert res.valid is True, f"err: {[e.as_dict() for e in res.errors]}"
    else:
        assert res.valid is False
        testutil.assert_jsonschema_error_contains(res, expected_err, expected_num_errs=1)


@pytest.fixture(name="fake_inputs")
def make_fake_inputs(tmp_path):
    fake_input_tree = make_fake_input_tree(tmp_path, {
        "/file1": "A file to be tarred",
        "/file2": "Another file to be tarred",
    })
    inputs = {
        "tree": {
            "path": fake_input_tree,
        }
    }
    return inputs


@pytest.mark.skipif(not has_executable("tar"), reason="no tar executable")
def test_tar_filename(tmp_path, stage_module, fake_inputs):
    options = {
        "filename": "out.tar",
    }
    stage_module.main(fake_inputs, tmp_path, options)

    tar_path = os.path.join(tmp_path, "out.tar")
    assert os.path.exists(tar_path)
    output = subprocess.check_output(["tar", "-tf", tar_path], encoding="utf-8").strip().split("\n")
    assert sorted(["./", "./file1", "./file2"]) == sorted(output)


@pytest.mark.skipif(not has_executable("tar"), reason="no tar executable")
def test_tar_root_node_omit(tmp_path, stage_module, fake_inputs):
    options = {
        "filename": "out.tar",
        "root-node": "omit",
    }
    stage_module.main(fake_inputs, tmp_path, options)

    tar_path = os.path.join(tmp_path, "out.tar")
    assert os.path.exists(tar_path)
    output = subprocess.check_output(["tar", "-tf", tar_path], encoding="utf-8").strip().split("\n")
    assert sorted(["file1", "file2"]) == sorted(output)


@pytest.mark.skipif(not has_executable("tar"), reason="no tar executable")
def test_tar_paths(tmp_path, stage_module, fake_inputs):
    options = {
        "filename": "out.tar",
        "paths": [
            "file2",
            "file1",
            "file1",
            "file2",
        ]
    }
    stage_module.main(fake_inputs, tmp_path, options)

    tar_path = os.path.join(tmp_path, "out.tar")
    assert os.path.exists(tar_path)
    output = subprocess.check_output(["tar", "-tf", tar_path], encoding="utf-8").split("\n")
    assert ["file2", "file1", "file1", "file2", ""] == output


@pytest.mark.skipif(not has_executable("tar"), reason="no tar executable")
@pytest.mark.parametrize("transform,expected_tar_output", [
    # unrelated transform
    ("s/foo/bar/", ["file1", "file2", ""]),
    # one file
    ("s/^file1$/foo99/", ["foo99", "file2", ""]),
    # one file
    ("s/file1/foo99/;s/file2/bar11/", ["foo99", "bar11", ""]),
])
def test_tar_transform(tmp_path, stage_module, fake_inputs, transform, expected_tar_output):
    options = {
        "filename": "out.tar",
        "paths": [
            "file1",
            "file2",
        ],
        "transform": transform,
    }
    stage_module.main(fake_inputs, tmp_path, options)

    tar_path = os.path.join(tmp_path, "out.tar")
    assert os.path.exists(tar_path)
    output = subprocess.check_output(["tar", "-tf", tar_path], encoding="utf-8").split("\n")
    assert output == expected_tar_output


def test_tar_transform_no_sh(tmp_path, stage_module, fake_inputs):
    options = {
        "filename": "out.tar",
        "paths": [
            "file1",
        ],
        # GNU sed allows to run shell commands with "/e"
        # ensure here we donot allow this
        "transform": "s/file1/date/e",
    }
    with pytest.raises(subprocess.CalledProcessError) as ex:
        stage_module.main(fake_inputs, tmp_path, options)
    assert "exit status 2" in str(ex.value)


@pytest.mark.parametrize('tmp_path_disk_full_size', [5 * 1024])
@pytest.mark.skipif(os.getuid() != 0, reason="needs root")
def test_tar_disk_full(stage_module, fake_inputs, tmp_path_disk_full, capfd):
    options = {
        "filename": "out.tar",
    }

    with pytest.raises(subprocess.CalledProcessError) as ex:
        stage_module.main(fake_inputs, tmp_path_disk_full, options)

    assert ex.value.returncode == 2
    assert re.search(r"Wrote only \d+ of \d+ bytes", str(capfd.readouterr().err))


@pytest.mark.skipif(not has_executable("tar"), reason="no tar executable")
@pytest.mark.parametrize("filename,compression,expected", [
    ("out.tar.gz", "auto", "gzip"),  # Auto uses the filename, which ends in .gz and should lead to gzip
    ("out.unknown", "xz", "xz"),
    ("out.unknown", "gzip", "gzip"),
    ("out.unknown", "zstd", "zstandard"),
])
def test_tar_compress(tmp_path, stage_module, fake_inputs, filename, compression, expected):
    options = {
        "filename": filename,
        "paths": [
            "file1",
            "file2",
        ],
        "compression": compression,
    }
    stage_module.main(fake_inputs, tmp_path, options)

    tar_path = os.path.join(tmp_path, filename)
    assert os.path.exists(tar_path)
    output = subprocess.check_output(["file", tar_path], encoding="utf-8")
    assert expected in output.lower()


@pytest.mark.skipif(not has_executable("tar"), reason="no tar executable")
@pytest.mark.parametrize("numeric_owner", [True, False,])
def test_tar_numeric_owner(tmp_path, stage_module, fake_inputs, numeric_owner):
    options = {
        "filename": "out.tar",
        "numeric-owner": numeric_owner,
    }
    stage_module.main(fake_inputs, tmp_path, options)

    tar_path = os.path.join(tmp_path, "out.tar")
    assert os.path.exists(tar_path)

    # read the tar archive directly instead of relying on parsing the human-readable verbose output
    with tarfile.open(tar_path, mode="r") as archive:
        for item in archive.getmembers():
            # when enabling numeric-owner, the uname and gname fields are empty
            info = item.get_info()
            assert (info["uname"] == "") == numeric_owner
            assert (info["gname"] == "") == numeric_owner



#!/usr/bin/python3

import os
import textwrap

import pytest

from osbuild import testutil

STAGE_NAME = "org.osbuild.containers.unit.create"


@pytest.mark.parametrize("test_data,expected_err", [
    # good
    (
        {
            "filename": "foo.container",
            "config": {
                "Unit": {},
                "Service": {},
                "Container": {
                    "Image": "img"
                },
                "Install": {},
            },
        },
        "",
    ),
    (
        {
            "filename": "foo.volume",
            "config": {
                "Unit": {},
                "Volume": {},
                "Install": {},
            },
        },
        "",
    ),
    (
        {
            "filename": "foo.network",
            "config": {
                "Network": {},
            },
        },
        "",
    ),
    # bad
    ({"config": {"Unit": {}, "Container": {"Image": "img"}, "Install": {}}}, "'filename' is a required property"),
    ({"filename": "foo.container"}, "'config' is a required property"),
    ({"filename": "foo.container", "config": {"Container": {"Image": "img"}, "Install": {}}},
     "{'Container': {'Image': 'img'}, 'Install': {}} is not valid under any of the given schemas"),
    ({"filename": "foo.container", "config": {"Unit": {}, "Install": {}}},
     "{'Unit': {}, 'Install': {}} is not valid under any of the given schemas"),
    ({"filename": "foo.container", "config": {"Network": {}, "Volume": {}}},
     "{'Network': {}, 'Volume': {}} is not valid under any of the given schemas")
])
@pytest.mark.parametrize("stage_schema", ["1"], indirect=True)
def test_schema_validation(stage_schema, test_data, expected_err):
    test_input = {
        "name": STAGE_NAME,
        "options": test_data,
    }
    res = stage_schema.validate(test_input)
    if expected_err == "":
        assert res.valid is True, f"err: {[e.as_dict() for e in res.errors]}"
    else:
        assert res.valid is False
        testutil.assert_jsonschema_error_contains(res, expected_err, expected_num_errs=1)


@pytest.mark.parametrize("test_data,expected_err", [
    # good
    (
        {
            "filename": "foo.service",
            "config": {
                "Unit": {},
                "Service": {},
                "Install": {},
            },
        },
        "",
    ),
    (
        {
            "filename": "foo.network",
            "config": {
                "Unit": {},
                "Network": {},
                "Install": {},
            },
        },
        "",
    ),
    (
        {
            "filename": "foo.volume",
            "config": {
                "Volume": {},
            },
        },
        "",
    ),
    # bad
    ({"filename": "something.container", "config": {"Unit": {}, "Volume": {}, "Install": {}}},
     "Error: something.container unit requires Container section"),
    ({"filename": "data-gifs-cats.volume", "config": {"Unit": {}, "Network": {}, "Install": {}}},
     "Error: data-gifs-cats.volume unit requires Volume section"),
])
def test_name_config_match(tmp_path, stage_module, test_data, expected_err):
    expected_unit_path = tmp_path / "usr/share/containers/systemd"
    expected_unit_path.mkdir(parents=True, exist_ok=True)
    try:
        stage_module.main(tmp_path, test_data)
    except ValueError as ve:
        assert expected_err == str(ve)


@pytest.mark.parametrize("unit_path,expected_prefix", [
    ("usr", "usr/share/containers/systemd"),
    ("etc", "etc/containers/systemd")
])
def test_systemd_unit_create(tmp_path, stage_module, unit_path, expected_prefix):
    options = {
        "filename": "create-directory.container",
        "unit-path": unit_path,
        "config": {
            "Unit": {
                "Description": "Create directory",
                "DefaultDependencies": False,
                "ConditionPathExists": [
                    "|!/etc/myfile"
                ],
                "ConditionPathIsDirectory": [
                    "|!/etc/mydir"
                ],
                # need to use real units otherwise the validation will fail
                "Wants": [
                    "basic.target",
                    "sysinit.target",
                ],
                "Requires": [
                    "paths.target",
                    "sockets.target",
                ],
                "After": [
                    "sysinit.target",
                    "default.target",
                ],
                "Before": [
                    "multi-user.target",
                ],
            },
            "Container": {
                "Image": "img",
                "Environment": [
                    {
                        "key": "DEBUG",
                        "value": "1",
                    },
                    {
                        "key": "TRACE",
                        "value": "1",
                    },
                ],
                "SecurityLabelFileType": "usr_t",
                "SecurityLabelType": "spc_t",
                "Tmpfs": "/test.tmp",
            },
            "Install": {
                "WantedBy": [
                    "local-fs.target"
                ],
                "RequiredBy": [
                    "multi-user.target"
                ]
            }
        }
    }

    expected_unit_path = tmp_path / expected_prefix / "create-directory.container"
    expected_unit_path.parent.mkdir(parents=True, exist_ok=True)

    stage_module.main(tmp_path, options)
    assert os.path.exists(expected_unit_path)
    assert expected_unit_path.read_text(encoding="utf-8") == textwrap.dedent("""\
    [Unit]
    Description=Create directory
    DefaultDependencies=False
    ConditionPathExists=|!/etc/myfile
    ConditionPathIsDirectory=|!/etc/mydir
    Wants=basic.target
    Wants=sysinit.target
    Requires=paths.target
    Requires=sockets.target
    After=sysinit.target
    After=default.target
    Before=multi-user.target

    [Container]
    Image=img
    Environment="DEBUG=1"
    Environment="TRACE=1"
    SecurityLabelFileType=usr_t
    SecurityLabelType=spc_t
    Tmpfs=/test.tmp

    [Install]
    WantedBy=local-fs.target
    RequiredBy=multi-user.target

    """)



#!/usr/bin/python3

import os.path

from osbuild.testutil import make_fake_tree

STAGE_NAME = "org.osbuild.ostree.grub2"


def test_ostree_grub2_boot(tmp_path, stage_module):
    make_fake_tree(tmp_path, {
        "ostree/boot.1/default/b6cdf47cafd171e003c001802f1829ad39e20f7177d6e27401aba73efb71be22/0": "",
        "grub.cfg": "linux /vmlinux ostree=@OSTREE@"
    })

    inputs = {
        "tree": {
            "path": tmp_path
        }
    }

    options = {
        "filename": "grub.cfg"
    }
    stage_module.main({
        "inputs": inputs,
        "tree": tmp_path,
        "options": options,
    })

    confpath = tmp_path / "grub.cfg"
    assert os.path.exists(confpath)
    assert confpath.read_text() == "linux /vmlinux ostree=/ostree/boot.1/default/b6cdf47cafd171e003c001802f1829ad39e20f7177d6e27401aba73efb71be22/0"



import abc
import hashlib
import json
import os
import shutil
import tempfile
from typing import ClassVar, Dict

from . import host
from .objectstore import ObjectStore


class SourceItemResult:
    def __init__(self, success=True, metadata=None, error=""):
        self.success = success
        self.metadata = metadata if metadata is not None else {}
        self.error = error

    def set_error(self, error):
        self.error = error
        self.success = False

    def as_dict(self):
        return {
            "success": self.success,
            "metadata": self.metadata,
            "error": self.error,
        }

    @classmethod
    def from_dict(cls, data):
        r = cls(data["success"])
        r.metadata = data.get("metadata", {})
        r.error = data.get("error", "")
        return r


class SourceResults:
    """Contains the results of a single source type.

    Because this is passed to monitor.BaseMonitor.result, the name, id,
    success and output properties need to be present.
    """

    def __init__(self, source_type):
        self.source_type = source_type
        self.success = True
        self.results = []

    def add(self, result: SourceItemResult):
        self.results.append(result)
        if not result.success:
            self.success = False

    @property
    def id(self):
        return self.source_type

    @property
    def name(self):
        return f"source {self.source_type}"

    @property
    def output(self):
        if self.success:
            return f"source {self.source_type} finished successfully"
        error = self.results[-1].error
        return f"source {self.source_type} finished unsuccessfully: {error}"

    def as_dict(self):
        return {
            "name": self.name,
            "source_type": self.source_type,
            "success": self.success,
            "output": self.output,
            "results": [r.as_dict() for r in self.results],
        }

    @classmethod
    def from_dict(cls, data):
        sr = cls(data["source_type"])
        sr.success = data["success"]
        sr.results = [SourceItemResult.from_dict(r) for r in data["results"]]
        return sr


class Source:
    """
    A single source with is corresponding options.
    """

    def __init__(self, info, items, options) -> None:
        self.info_name = info.name
        self.info_path = info.path
        self.items = items or {}
        self.options = options
        # compat with pipeline
        self.build = None
        self.runner = None
        self.source_epoch = None

    def download(self, mgr: host.ServiceManager, store: ObjectStore) -> SourceResults:
        source = self.info_name
        cache = os.path.join(store.store, "sources")

        args = {
            "items": self.items,
            "options": self.options,
            "cache": cache,
            "output": None,
            "checksums": [],
        }

        client = mgr.start(f"source/{source}", self.info_path)
        reply = client.call("download", args)
        return SourceResults.from_dict(reply)

    def copy(self, mgr: host.ServiceManager, store: ObjectStore, dst_store: ObjectStore) -> SourceResults:
        source = self.info_name
        cache = os.path.join(store.store, "sources")
        dst_cache = os.path.join(dst_store.store, "sources")

        args = {
            "items": self.items,
            "options": self.options,
            "cache": cache,
            "dst_cache": dst_cache,
        }

        client = mgr.start(f"source/{source}", self.info_path)
        reply = client.call("copy", args)
        return SourceResults.from_dict(reply)

    # "name", "id", "stages", "results" is only here to make it looks like a
    # pipeline for the monitor. This should be revisited at some point
    # and maybe the monitor should get first-class support for
    # sources?
    #
    # In any case, sources can be represented only poorly right now
    # by the monitor because the source is called with download()
    # for all items and there is no way for a stage right now to
    # report something structured back to the host that runs the
    # source so it just downloads all sources without any user
    # visible progress right now
    @property
    def name(self):
        return f"source {self.info_name}"

    @property
    def id(self):
        m = hashlib.sha256()
        m.update(json.dumps(self.info_name, sort_keys=True).encode())
        m.update(json.dumps(self.items, sort_keys=True).encode())
        return m.hexdigest()

    @property
    def stages(self):
        return []


class SourceService(host.Service):
    """Source host service"""

    max_workers = 1

    content_type: ClassVar[str]
    """The content type of the source."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.cache = None
        self.options = None
        self.tmpdir = None

    @abc.abstractmethod
    def fetch_one(self, checksum, desc) -> SourceItemResult:
        """Performs the actual fetch of an element described by its checksum and its descriptor"""

    @abc.abstractmethod
    def fetch_all(self, items: Dict) -> Dict:
        """Fetch all sources."""

    def exists(self, checksum, _desc) -> bool:
        """Returns True if the item to download is in cache. """
        return os.path.isfile(f"{self.cache}/{checksum}")

    def copy_all(self, items: Dict, dst_cache: str) -> Dict:
        """Copies the source to the destination"""
        target_dir = os.path.join(dst_cache, self.content_type)
        os.makedirs(target_dir, exist_ok=True)
        results = SourceResults(self.id)
        for item in items:
            res = SourceItemResult(metadata={
                "checksum": item,
            })
            try:
                shutil.copy2(os.path.join(self.cache, item), os.path.join(target_dir, item))
            except Exception as e:  # pylint: disable=broad-exception-caught
                res.set_error(str(e))
            results.add(res)
        return results.as_dict()

    def setup(self, args):
        self.cache = os.path.join(args["cache"], self.content_type)
        os.makedirs(self.cache, exist_ok=True)
        self.options = args["options"]

    def dispatch(self, method: str, args, fds):
        if method == "download":
            self.setup(args)
            with tempfile.TemporaryDirectory(prefix=".unverified-", dir=self.cache) as self.tmpdir:
                sourceresults = self.fetch_all(args["items"])
                return sourceresults, None

        elif method == "copy":
            self.setup(args)
            sourceresults = self.copy_all(args["items"], args["dst_cache"])
            return sourceresults, None

        raise host.ProtocolError("Unknown method")



#!/usr/bin/python3
from unittest import mock

import pytest

from osbuild import testutil

STAGE_NAME = "org.osbuild.bootctl.install.root"


@pytest.mark.parametrize("test_data,expected_err", [
    # bad
    ({}, "'root' is a required property"),
    ({"root": "/foo/bar"}, "'/foo/bar' is not valid under any of the given schemas"),
    ({"root": "unknown://-/"}, "'unknown://-/' is not valid under any of the given schemas"),
    ({"root": "mount://-/", "unknown": "property"}, "Additional properties are not allowed ('unknown' was unexpected)"),
    ({"root": "input://-/"}, "'input://-/' is not valid under any of the given schemas"),
    ({"root": "tree:///", "random-seed": "false"}, "'false' is not one of ['yes', 'no']"),
    ({"root": "tree:///", "make-entry-directory": "false"}, "'false' is not one of ['yes', 'no']"),
    ({"root": "tree:///", "entry-token": "broken"},
     "'broken' is not one of ['auto', 'machine-id', 'os-id', 'os-image-id']"),
    # good
    ({"root": "mount://-/"}, ""),
    ({"root": "tree:///"}, ""),
    ({"root": "tree:///with/subpath"}, ""),
    ({"root": "mount://-/", "esp-path": "/foo/bar"}, ""),
    ({"root": "mount://-/", "boot-path": "/foo/bar"}, ""),
    ({"root": "mount://-/", "random-seed": "no"}, ""),
    ({"root": "mount://-/", "make-entry-directory": "yes"}, ""),
    ({"root": "mount://-/", "entry-token": "os-id"}, ""),
])
def test_bootctl_install_root_schema_validation(stage_schema, test_data, expected_err):
    test_input = {
        "type": STAGE_NAME,
        "options": {
        }
    }
    test_input["options"].update(test_data)
    res = stage_schema.validate(test_input)

    if expected_err == "":
        assert res.valid is True, f"err: {[e.as_dict() for e in res.errors]}"
    else:
        assert res.valid is False
        testutil.assert_jsonschema_error_contains(res, expected_err, expected_num_errs=1)


@mock.patch("subprocess.run")
def test_bootctl_install_root_plain(mock_run, tmp_path, stage_module):
    options = {
        "root": "tree:///",
    }

    fake_tree = tmp_path / "tree"

    stage_module.main({
        "tree": fake_tree,
        "options": options,
    })

    mock_run.assert_called_with([
        "bootctl", "install",
        "--no-variables",
        "--root", str(fake_tree),
    ],
        check=True,
        env={},
    )


@mock.patch("subprocess.run")
def test_bootctl_install_root_with_options(mock_run, tmp_path, stage_module):
    options = {
        "root": "tree:///",
        "esp-path": "/efi",
        "boot-path": "/boot",
        "relax-esp-checks": True,
        "random-seed": "no",
        "make-entry-directory": "yes",
        "entry-token": "os-id"
    }

    fake_tree = tmp_path / "tree"

    stage_module.main({
        "tree": fake_tree,
        "options": options,
    })

    mock_run.assert_called_with([
        "bootctl", "install",
        "--no-variables",
        "--root", str(fake_tree),
        "--random-seed", "no",
        "--make-entry-directory", "yes",
        "--entry-token", "os-id",
        "--esp-path", "/efi",
        "--boot-path", "/boot",
    ],
        check=True,
        env={"SYSTEMD_RELAX_ESP_CHECKS": 1},
    )



#!/usr/bin/python3

import os

SOURCES_NAME = "org.osbuild.skopeo"


def test_skopeo_copy_all(tmp_path, sources_service):
    src_cache = tmp_path / "src-cache"
    dst_cache = tmp_path / "dst-cache"
    os.mkdir(src_cache)
    os.mkdir(dst_cache)

    sources_service.setup({"cache": src_cache, "options": {}})

    # simulate a container in the store
    checksum = "sha256:ca978112ca1bbdcafac231b39a23dc4da786eff8147c4e72b9807785afee48bb"
    os.makedirs((src_cache / sources_service.content_type / checksum / "image"))
    (src_cache / sources_service.content_type / checksum / "image" / "version").write_text("Directory Transport Version: 1.1")
    (src_cache / sources_service.content_type / checksum /
     "image" / "manifest.json").write_text("{\"schemaVersion\":2}")
    (src_cache / sources_service.content_type / checksum / "image" /
     "1111111111111111111111111111111111111111111111111111111111111111").write_text("1")
    (src_cache / sources_service.content_type / checksum / "image" /
     "2222222222222222222222222222222222222222222222222222222222222222").write_text("2")

    assert sources_service.exists(checksum, None)
    sources_service.copy_all({checksum: {}}, dst_cache)
    assert os.path.isdir(dst_cache / sources_service.content_type / checksum / "image")
    assert (dst_cache / sources_service.content_type / checksum / "image" /
            "version").read_text() == "Directory Transport Version: 1.1"
    assert (dst_cache / sources_service.content_type / checksum / "image" /
            "manifest.json").read_text() == "{\"schemaVersion\":2}"
    assert (dst_cache / sources_service.content_type / checksum / "image" /
            "1111111111111111111111111111111111111111111111111111111111111111").read_text() == "1"
    assert (dst_cache / sources_service.content_type / checksum / "image" /
            "2222222222222222222222222222222222222222222222222222222222222222").read_text() == "2"



#!/usr/bin/python3

import gzip
import os
from unittest import mock

import pytest

from osbuild import testutil
from osbuild.testutil import has_executable

STAGE_NAME = "org.osbuild.gunzip"


# Prepare dataset containing good and bad API call parameters
@pytest.mark.parametrize("test_data, expected_err", [
    # Bad API parameters
    ({}, "'path' is a required property"),
    ({"path": 123}, "123 is not of type 'string'"),
    # Good API parameters
    ({"path": "/tmp/"}, ""),
    ({"path": "/boot/initramfs-5.14.0-247.el9.x86_64.img"}, ""),
])
# This test validates only API calls using correct and incorrect queries
def test_schema_validation_gunzip(stage_schema, test_data, expected_err):
    test_input = {
        "type": STAGE_NAME,
        "options": {
        }
    }

    test_input["options"].update(test_data)
    res = stage_schema.validate(test_input)

    if expected_err == "":
        assert res.valid is True, f"err: {[e.as_dict() for e in res.errors]}"
    else:
        assert res.valid is False
        testutil.assert_jsonschema_error_contains(res, expected_err, expected_num_errs=1)


# This test creates a compressed dummy file then decompresses and
# compares the content.
@pytest.mark.skipif(not has_executable("gunzip"), reason="gunzip")
def test_gunzip_integration(tmp_path, stage_module):
    fake_tree = tmp_path / "fake/output/tree"
    fake_tree.mkdir(parents=True)
    fake_archive_path = fake_tree / "some-file.gz"
    fake_file_content = b"Some random text written to a file"

    with gzip.GzipFile(fake_archive_path, "w") as fp:
        fp.write(fake_file_content)

    inp = {
        "file": {
            "path": fake_tree,
            "data": {
                "files": {
                    fake_archive_path.name: "Some test file"
                }
            },
        },
    }
    expected_output_path = fake_tree / "tree" / "output.txt"
    expected_output_path.parent.mkdir()
    options = {
        "path": expected_output_path.name
    }
    stage_module.main(inp, expected_output_path.parent, options)
    assert expected_output_path.read_bytes() == fake_file_content


# This test evaluates final binary call and match between its flags and stage args
@mock.patch("subprocess.run")
def test_gunzip_cmdline(mock_run, tmp_path, stage_module):
    fake_tree = tmp_path / "fake/output/tree"
    fake_tree.mkdir(parents=True)
    fake_file = "fake_file.txt"
    fake_archive = f"{fake_file}.gz"

    inp = {
        "file": {
            "path": fake_tree,
            "data": {
                "files": {
                    fake_archive: "Some test file"
                }
            },
        },
    }
    options = {
        "path": fake_file
    }
    output = fake_tree
    stage_module.main(inp, output, options)

    expected = [
        "gunzip",
        "--stdout",
    ] + [os.fspath(fake_tree / fake_archive)]
    mock_run.assert_called_once_with(expected, stdout=mock.ANY, check=True)



#!/usr/bin/python3

import os

import pytest

from osbuild import testutil

STAGE_NAME = "org.osbuild.systemd-sysusers.d"


@pytest.mark.parametrize(
    "test_data",
    [
        (
            {
                "config": {
                    "filename": "new.conf",
                    "entries": [
                        {
                            "type": "u",
                            "name": "test-user",
                        }
                    ],
                }
            },
        ),
        (
            {
                "config": {
                    "filename": "new.conf",
                    "entries": [
                        {
                            "type": "u!",
                            "name": "test-user",
                        }
                    ],
                }
            },
        ),
        (
            {
                "config": {
                    "filename": "new.conf",
                    "entries": [
                        {
                            "type": "g",
                            "name": "test-user",
                        }
                    ],
                }
            },
        ),
        (
            {
                "config": {
                    "filename": "new.conf",
                    "entries": [
                        {
                            "type": "m",
                            "name": "test-user",
                        }
                    ],
                }
            },
        ),
        (
            {
                "config": {
                    "filename": "new.conf",
                    "entries": [
                        {
                            "type": "r",
                            "name": "test-user",
                        }
                    ],
                }
            },
        ),
        (
            {
                "config": {
                    "filename": "new.conf",
                    "entries": [
                        {
                            "type": "u",
                            "name": "very-long-user_with_number-1234",
                        }
                    ],
                }
            },
        ),
        (
            {
                "config": {
                    "filename": "new.conf",
                    "entries": [
                        {
                            "type": "u",
                            "name": "very-long-user_with_number-1234",
                        }
                    ],
                },
            },
        ),
        # for type r the name needs to be a single dash symbol
        (
            {
                "config": {
                    "filename": "new.conf",
                    "entries": [{"type": "r", "name": "-"}],
                },
            },
        ),
    ],
)
@pytest.mark.parametrize("stage_schema", ["1"], indirect=True)
def test_schema_validation_good(stage_schema, test_data):
    test_input = {
        "name": STAGE_NAME,
        "options": test_data,
    }
    res = stage_schema.validate(test_input)
    assert res.valid is True, f"err: {[e.as_dict() for e in res.errors]}"


@pytest.mark.parametrize(
    "test_data,expected_err",
    [
        # required field 'filename' of config is missing
        (
            {
                "config": {"entries": []},
            },
            "'filename' is a required property",
        ),
        # required field 'entries' of config is missing
        (
            {
                "config": {"filename": "new.conf"},
            },
            "'entries' is a required property",
        ),
        # required field 'filename' is invalid
        (
            {
                "config": {"filename": "new.cof", "entries": []},
            },
            "'new.cof' does not match '\\\\.conf$'",
        ),
        # field 'path-prefix' is invalid
        (
            {
                "config": {"path-prefix": "run", "filename": "new.conf", "entries": []},
            },
            "'run' is not one of ['usr', 'etc']",
        ),
        # required field 'type' of entries is missing
        (
            {
                "config": {
                    "filename": "new.conf",
                    "entries": [
                        {
                            "type": "u",
                        }
                    ],
                },
            },
            "'name' is a required property",
        ),
        # required field 'type' of entries is missing
        (
            {
                "config": {
                    "filename": "new.conf",
                    "entries": [
                        {
                            "name": "test-user",
                        }
                    ],
                },
            },
            "'type' is a required property",
        ),
        # invalid field 'type' of entries
        (
            {
                "config": {"filename": "new.conf", "entries": [{"type": "x", "name": "test-user"}]},
            },
            "'x' is not one of ['u', 'u!', 'g', 'm', 'r']",
        ),
        # invalid field 'name' of entries - too long
        (
            {
                "config": {
                    "filename": "new.conf",
                    "entries": [{"type": "u", "name": "too-long-user_with_number-1234-and-more"}],
                },
            },
            "'too-long-user_with_number-1234-and-more' does not match '^([a-z]+(?:[-_][a-zA-Z0-9_-]{0,30})?|-)$",
        ),
        # invalid field 'name' of entries - starts with dash symbol
        (
            {
                "config": {
                    "filename": "new.conf",
                    "entries": [{"type": "u", "name": "-also-invalid"}],
                },
            },
            "'-also-invalid' does not match '^([a-z]+(?:[-_][a-zA-Z0-9_-]{0,30})?|-)$",
        ),
    ],
)
@pytest.mark.parametrize("stage_schema", ["1"], indirect=True)
def test_schema_validation_bad(stage_schema, test_data, expected_err):
    test_input = {
        "name": STAGE_NAME,
        "options": test_data,
    }
    res = stage_schema.validate(test_input)
    assert res.valid is False
    testutil.assert_jsonschema_error_contains(res, expected_err, expected_num_errs=1)


@pytest.mark.parametrize(
    "test_config,expected_content",
    [
        (
            {
                "filename": "custom.conf",
                "entries": [
                    {
                        "type": "u",
                        "name": "sysuser-a",
                    }
                ],
            },
            "u sysuser-a - - - -\n",
        ),
        (
            {
                "path-prefix": "etc",
                "filename": "custom.conf",
                "entries": [
                    {
                        "type": "u",
                        "name": "sysuser-a",
                    }
                ],
            },
            "u sysuser-a - - - -\n",
        ),
        (
            {
                "filename": "custom.conf",
                "entries": [{"type": "u", "name": "sysuser-a", "id": "-"}],
            },
            "u sysuser-a - - - -\n",
        ),
        (
            {
                "filename": "custom.conf",
                "entries": [
                    {
                        "type": "u",
                        "name": "sysuser-a",
                        "id": "100",
                        "gecos": "Test user for the system",
                        "home": "/home/sysuser",
                        "shell": "/bin/bash",
                    }
                ],
            },
            "u sysuser-a 100 \"Test user for the system\" /home/sysuser /bin/bash\n",
        ),
        (
            {
                "filename": "custom.conf",
                "entries": [
                    {
                        "type": "u",
                        "name": "sysuser-a",
                        "id": "100:200",
                    }
                ],
            },
            "u sysuser-a 100:200 - - -\n",
        ),
        (
            {
                "filename": "custom.conf",
                "entries": [
                    {
                        "type": "u",
                        "name": "sysuser-a",
                        "id": "100",
                        "gecos": "Test user for the system",
                        "home": "/home/sysuser",
                        "shell": "/bin/bash",
                    },
                    {
                        "type": "g",
                        "name": "sysuser-a-group",
                        "id": "101",
                    },
                    {
                        "type": "m",
                        "name": "sysuser-a",
                        "id": "sysuser-a-group",
                    },
                    {
                        "type": "r",
                        "name": "-",
                        "id": "200-500",
                    },
                ],
            },
            """u sysuser-a 100 \"Test user for the system\" /home/sysuser /bin/bash
g sysuser-a-group 101 - - -
m sysuser-a sysuser-a-group - - -
r - 200-500 - - -
""",
        ),
        (
            {
                "filename": "custom.conf",
                "entries": [
                    {
                        "type": "u",
                        "name": "sysuser-a",
                        "home": "/home/sysuser",
                        "shell": "/bin/bash",
                    },
                ],
            },
            """u sysuser-a - - /home/sysuser /bin/bash
""",
        ),
        (
            {
                "filename": "custom.conf",
                "entries": [
                    {
                        "type": "u",
                        "name": "sysuser-a",
                        "shell": "/bin/bash",
                    },
                ],
            },
            """u sysuser-a - - - /bin/bash
""",
        ),
    ]
)
def test_systemd_sysusers_d_config(tmp_path, stage_module, test_config, expected_content):
    options = {
        "config": test_config,
    }

    sysusers_d_dir = "usr/lib/sysusers.d"
    if "path-prefix" in test_config and test_config.get("path-prefix") == "etc":
        sysusers_d_dir = "etc/sysusers.d"
    os.makedirs(tmp_path / sysusers_d_dir)

    stage_module.main({"tree": tmp_path, "options": options})

    if test_config and "filename" in test_config:
        expected_conf_path = tmp_path / sysusers_d_dir / test_config["filename"]
        if "entries" in test_config and test_config["entries"]:
            assert os.path.exists(expected_conf_path)
            assert expected_conf_path.read_text(encoding="utf-8") == expected_content
        else:
            assert not os.path.exists(expected_conf_path)


@pytest.mark.parametrize(
    "test_config,expected_error",
    [
        # config is required
        (
            None,
            "'config' is required",
        ),
        # checking for missing required fields
        (
            {},
            "'filename' is required",
        ),
        (
            {
                "filename": "custom.conf",
            },
            "'entries' is required",
        ),
        (
            {
                "entries": {},
            },
            "'filename' is required",
        ),
        (
            {
                "filename": "custom.conf",
                "entries": [{}],
            },
            "Error: sysuser entry 'Type=' is invalid: type needs to be one of: ['u', 'u!', 'g', 'm', 'r']",
        ),
        (
            {
                "filename": "custom.conf",
                "entries": [
                    {
                        "type": "u",
                    }
                ],
            },
            "Error: sysuser entry 'Name=' is invalid: name needs to match regex '^([a-z]+(?:[-_][a-zA-Z0-9_-]{0,30})?|-)$'",
        ),
        # checking name validation
        (
            {
                "filename": "custom.conf",
                "entries": [
                    {
                        "type": "u",
                        "name": "sysuser-a-with-a-very-very-long-id-123456",
                    },
                ],
            },
            "Error: sysuser entry 'Name=sysuser-a-with-a-very-very-long-id-123456' is invalid: name needs to match regex '^([a-z]+(?:[-_][a-zA-Z0-9_-]{0,30})?|-)$'",
        ),
        # checking id validation
        (
            {
                "filename": "custom.conf",
                "entries": [
                    {"type": "u", "name": "sysuser-a", "id": "blubb"},
                ],
            },
            "invalid literal for int() with base 10: 'blubb'",
        ),
        (
            {
                "filename": "custom.conf",
                "entries": [
                    {
                        "type": "u",
                        "name": "sysuser-a",
                        "id": "100-200",
                    }
                ],
            },
            "invalid literal for int() with base 10: '100-200'",
        ),
        (
            {
                "filename": "custom.conf",
                "entries": [
                    {
                        "type": "r",
                        "name": "-",
                        "id": "200:500",
                    }
                ],
            },
            "invalid literal for int() with base 10: '200:500'",
        ),
        (
            {
                "filename": "custom.conf",
                "entries": [
                    {
                        "type": "r",
                        "name": "a-name",
                        "id": "200-500",
                    }
                ],
            },
            "Error: sysuser entry 'Name=a-name' is invalid: name needs to be '-' for type 'r'",
        ),
        (
            {
                "filename": "custom.conf",
                "entries": [
                    {
                        "type": "m",
                        "name": "a-user",
                        "id": "101",
                    }
                ],
            },
            "Error: sysuser entry 'ID=101' is invalid: id should be group name for m entry",
        ),
    ],
)
def test_systemd_sysusers_d_config_bad(tmp_path, stage_module, test_config, expected_error):
    options = {
        "config": test_config,
    }

    sysusers_d_dir = "etc/sysusers.d"
    os.makedirs(tmp_path / sysusers_d_dir)

    with pytest.raises(ValueError) as ex:
        stage_module.main({"tree": tmp_path, "options": options})
    assert expected_error in str(ex.value)



#!/usr/bin/python3

import pytest

from osbuild.testutil import assert_jsonschema_error_contains

STAGE_NAME = "org.osbuild.groups"


@pytest.mark.parametrize("test_data,expected_err", [
    # bad
    ({"groups": {"!invalid-name": {}}}, "'!invalid-name' does not match any of the regex"),
    ({"groups": {"foo": {"gid": "x"}}}, "'x' is not of type 'number'"),
    # good
    ({}, ""),
    ({"groups": {"foo": {}}}, ""),
    ({"groups": {"foo": {"gid": 999}}}, ""),
    ({"groups": {"foo": {"gid": 999}}, "force": True}, ""),
    ({"groups": {"foo.bar.baz": {}}}, ""),
])
def test_schema_validation(stage_schema, test_data, expected_err):
    test_input = {
        "type": STAGE_NAME,
        "options": {},
    }
    test_input["options"].update(test_data)
    res = stage_schema.validate(test_input)

    if expected_err == "":
        assert res.valid is True, f"err: {[e.as_dict() for e in res.errors]}"
    else:
        assert res.valid is False
        assert_jsonschema_error_contains(res, expected_err, expected_num_errs=1)



import os.path
import re

import pytest

from osbuild import testutil
from osbuild.util import toml

STAGE_NAME = "org.osbuild.bootc.install.config"


@pytest.mark.parametrize(
    "options,expected_config",
    [
        (
            {
                "filename": "10-ten.conf",
                "config": {
                    "install": {
                        "filesystem": {
                            "root": {
                                "type": "xfs",
                            }
                        }
                    }
                }
            },
            {
                "install": {
                    "filesystem": {
                        "root": {
                            "type": "xfs"
                        }
                    }
                }
            },
        ),
        (
            {
                "filename": "20-twenty.conf",
                "config": {
                    "install": {
                        "kargs": [
                            "ro",
                            "biosdevname=0",
                            "net.ifnames=0",
                        ]
                    }
                }
            },
            {
                "install": {
                    "kargs": ["ro", "biosdevname=0", "net.ifnames=0"],
                }
            },
        ),
        (
            {
                "filename": "error.conf",
                "config": {
                    "install": {
                        "block": [
                            "tmp2-luks",
                        ],
                    }
                },
            },
            {
                "install": {
                    "block": ["tmp2-luks"]
                }
            }
        ),
        (
            {
                "filename": "30-ostree.conf",
                "config": {
                    "install": {
                        "ostree": {
                            "bls-append-except-default": 'grub_users=""'
                        }
                    }
                }
            },
            {
                "install": {
                    "ostree": {
                        "bls-append-except-default": 'grub_users=""'
                    }
                }
            },
        ),
    ]
)
@pytest.mark.tomlwrite
def test_bootc_install_config(tmp_path, stage_module, options, expected_config):
    stage_module.main(tmp_path, options)
    # different toml libraries write out arrays in different ways (one line vs many), so we need to parse the result
    # to compare
    config_data = toml.load_from_file(os.path.join(tmp_path, "usr/lib/bootc/install", options["filename"]))
    assert config_data == expected_config


@pytest.mark.parametrize(
    "options,expected_errors",
    [
        (
            {},
            ["'config' is a required property", "'filename' is a required property"]
        ),
        (
            {
                "filename": "error.conf",
            },
            ["'config' is a required property"]
        ),
        (
            {
                "filename": "error.conf",
                "config": {},
            },
            # under py3.6 the message is
            #   {} does not have enough properties
            # under other versions the message is
            #   {} should be non-empty
            [re.compile(r"{} should be non-empty|{} does not have enough properties")]
        ),
        (
            {
                "filename": "error.conf",
                "config": {
                    "install": {}
                },
            },
            [re.compile(r"{} should be non-empty|{} does not have enough properties")]
        ),
        (
            {
                "filename": "error.conf",
                "config": {
                    "install": {
                        "filesystem": {}
                    }
                },
            },
            ["'root' is a required property"]
        ),
        (
            {
                "filename": "error.conf",
                "config": {
                    "install": {
                        "filesystem": {
                            "root": {}
                        }
                    }
                },
            },
            ["'type' is a required property"]
        ),
        (
            {
                "filename": "error.conf",
                "config": {
                    "install": {
                        "block": [
                            "whatever",
                        ],
                    }
                },
            },
            ["'whatever' is not one of ['direct', 'tmp2-luks']"]
        ),
        (
            {
                "filename": "error.conf",
                "config": {
                    "install": {
                        "ostree": {
                            "invalid-option": "value",
                        },
                    }
                },
            },
            ["Additional properties are not allowed ('invalid-option' was unexpected)"]
        ),
    ]
)
def test_schema_bootc_install_config(stage_schema, options, expected_errors):
    test_input = {
        "type": STAGE_NAME,
        "options": options,
    }
    res = stage_schema.validate(test_input)
    assert res.valid is False
    for err in expected_errors:
        testutil.assert_jsonschema_error_contains(res, err)



"""
Function for appending parameters to
Boot Loader Specification (BLS).
"""
import glob
import os
from typing import List


def options_append(root_path: str, kernel_arguments: List[str]) -> None:
    """
    Add kernel arguments to the Boot Loader Specification (BLS) configuration files.
    There is unlikely to be more than one BLS config, but just in case, we'll iterate over them.

    Parameters
    ----------

    root_path (str): The root path for locating BLS configuration files.
    kernel_arguments (list): A list of kernel arguments to be added.

    """
    bls_glob = f"{root_path}/loader/entries/*.conf"
    bls_conf_files = glob.glob(bls_glob)
    if len(bls_conf_files) == 0:
        raise RuntimeError(f"no BLS configuration found in {bls_glob}")
    for entry in bls_conf_files:
        with open(entry, encoding="utf8") as f:
            lines = f.read().splitlines()
        with open(entry + ".tmp", "w", encoding="utf8") as f:
            found_opts_line = False
            for line in lines:
                if not found_opts_line and line.startswith('options '):
                    f.write(f"{line} {' '.join(kernel_arguments)}\n")
                    found_opts_line = True
                else:
                    f.write(f"{line}\n")
            if not found_opts_line:
                f.write(f"options {' '.join(kernel_arguments)}\n")
        os.rename(entry + ".tmp", entry)



"""Mount utilities
"""

import contextlib
import enum
import subprocess
from typing import Optional


class MountPermissions(enum.Enum):
    READ_WRITE = "rw"
    READ_ONLY = "ro"


def mount(source, target, bind=True, ro=True, private=True, mode="0755"):
    options = []
    if ro:
        options += [MountPermissions.READ_ONLY.value]
    if mode:
        options += [mode]

    args = []
    if bind:
        args += ["--rbind"]
    if private:
        args += ["--make-rprivate"]
    if options:
        args += ["-o", ",".join(options)]

    r = subprocess.run(["mount"] + args + [source, target],
                       stderr=subprocess.STDOUT,
                       stdout=subprocess.PIPE,
                       encoding="utf-8",
                       check=False)

    if r.returncode != 0:
        code = r.returncode
        msg = r.stdout.strip()
        raise RuntimeError(f"{msg} (code: {code})")


def umount(target, lazy=False):
    args = []
    if lazy:
        args += ["--lazy"]
    subprocess.run(["sync", "-f", target], check=True)
    subprocess.run(["umount", "-R"] + args + [target], check=True)


class MountGuard(contextlib.AbstractContextManager):
    def __init__(self):
        self.mounts = []
        self.remount = False

    def mount(
            self,
            source,
            target,
            bind=True,
            remount=False,
            permissions: Optional[MountPermissions] = None,
            mode="0755"):
        self.remount = remount
        options = []
        if bind:
            options += ["bind"]
        if remount:
            options += ["remount"]
        if permissions:
            if permissions not in list(MountPermissions):
                raise ValueError(f"unknown filesystem permissions: {permissions}")
            options += [permissions.value]
        if mode:
            options += [mode]

        args = ["--make-private"]
        if options:
            args += ["-o", ",".join(options)]

        r = subprocess.run(["mount"] + args + [source, target],
                           stderr=subprocess.STDOUT,
                           stdout=subprocess.PIPE,
                           encoding="utf-8",
                           check=False)
        if r.returncode != 0:
            code = r.returncode
            msg = r.stdout.strip()
            raise RuntimeError(f"{msg} (code: {code})")

        self.mounts += [{"source": source, "target": target}]

    def umount(self):

        while self.mounts:
            mnt = self.mounts.pop()  # FILO: get the last mount
            target = mnt["target"]
            # The sync should in theory not be needed but in rare
            # cases `target is busy` error has been spotted.
            # Calling  `sync` does not hurt so we keep it for now.
            if not self.remount:
                subprocess.run(["sync", "-f", target], check=True)
                subprocess.run(["umount", target], check=True)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.umount()



#!/usr/bin/python3

import os.path
from unittest.mock import call, patch

import pytest

STAGE_NAME = "org.osbuild.grub2.iso"

CONFIG_PART_1 = """
function load_video {
  insmod efi_gop
  insmod efi_uga
  insmod video_bochs
  insmod video_cirrus
  insmod all_video
}

load_video
set gfxpayload=keep
insmod gzio
insmod part_gpt
insmod ext2

set timeout=60
### END /etc/grub.d/00_header ###

search --no-floppy --set=root -l 'Fedora-42-Everything-x86_64'

"""

CONFIG_TMPL_CUSTOM = """
menuentry '{name}' --class fedora --class gnu-linux --class gnu --class os {{
	linux {linux}
	initrd {initrd}
}}"""

CONFIG_PART_INSTALL = """
menuentry 'Install Fedora 42' --class fedora --class gnu-linux --class gnu --class os {
	linux /images/pxeboot/vmlinuz inst.stage2=hd:LABEL=Fedora-42-Everything-x86_64 quiet
	initrd /images/pxeboot/initrd.img
}
"""

CONFIG_PART_TEST = """
menuentry 'Test this media & install Fedora 42' --class fedora --class gnu-linux --class gnu --class os {
	linux /images/pxeboot/vmlinuz inst.stage2=hd:LABEL=Fedora-42-Everything-x86_64 rd.live.check quiet
	initrd /images/pxeboot/initrd.img
}
"""

CONFIG_PART_TROUBLESHOOTING = """

submenu 'Troubleshooting -->' {
	menuentry 'Install Fedora 42 in basic graphics mode' --class fedora --class gnu-linux --class gnu --class os {
		linux /images/pxeboot/vmlinuz inst.stage2=hd:LABEL=Fedora-42-Everything-x86_64 nomodeset quiet
		initrd /images/pxeboot/initrd.img
	}
	menuentry 'Rescue a Fedora system' --class fedora --class gnu-linux --class gnu --class os {
		linux /images/pxeboot/vmlinuz inst.stage2=hd:LABEL=Fedora-42-Everything-x86_64 inst.rescue quiet
		initrd /images/pxeboot/initrd.img
	}
}
"""

CONFIG_FIPS = """
menuentry 'Install Fedora 42 in FIPS mode' --class fedora --class gnu-linux --class gnu --class os {
	linux /images/pxeboot/vmlinuz inst.stage2=hd:LABEL=Fedora-42-Everything-x86_64 quiet fips=1
	initrd /images/pxeboot/initrd.img
}
"""

CONFIG_DEFAULT = """set default="1"
"""


@patch("shutil.copy2")
@pytest.mark.parametrize("test_data,expected_conf", [
    # default
    ({}, CONFIG_PART_1 + "\n" + CONFIG_PART_INSTALL + CONFIG_PART_TEST + CONFIG_PART_TROUBLESHOOTING),
    # fips menu enable
    ({"fips": True}, CONFIG_PART_1 + "\n" + CONFIG_PART_INSTALL +
     CONFIG_PART_TEST + CONFIG_FIPS + CONFIG_PART_TROUBLESHOOTING),
    # default to menu entry 1
    ({"config": {"default": 1}}, CONFIG_DEFAULT + CONFIG_PART_1 + "\n" +
     CONFIG_PART_INSTALL + CONFIG_PART_TEST + CONFIG_PART_TROUBLESHOOTING),
    # no troubleshooting
    ({"troubleshooting": False}, CONFIG_PART_1 + "\n" + CONFIG_PART_INSTALL + CONFIG_PART_TEST + "\n\n"),
    # only install
    ({"troubleshooting": False, "test": False}, CONFIG_PART_1 + "\n" + CONFIG_PART_INSTALL + "\n\n\n"),
    # nothing
    ({"troubleshooting": False, "test": False, "install": False}, CONFIG_PART_1 + "\n\n\n\n\n"),
    # custom entries
    (
        {
            "kernel": {
                "dir": "/images/pxeboot",
                "opts": [
                    "root=hd:LABEL=root",
                    "rhgb"
                ]
            },
            "troubleshooting": False,
            "test": False,
            "install": False,
            "custom": [
                {
                    "name": "label 0",
                    "linux": "/foo/bar root=baz quiet",
                    "initrd": "/foo/bar",
                },
                {
                    "name": "label 1",
                    "linux": "/bar/foo root=baz quiet",
                    "initrd": "/bar/foo",
                },
                {
                    "name": "label 2",
                    "linux": "${kernelpath} ${root}",
                    "initrd": "${initrdpath}",
                },
            ]
        },
        CONFIG_PART_1 +
        CONFIG_TMPL_CUSTOM.format(
            name="label 0",
            linux="/foo/bar root=baz quiet",
            initrd="/foo/bar",
        ) +
        CONFIG_TMPL_CUSTOM.format(
            name="label 1",
            linux="/bar/foo root=baz quiet",
            initrd="/bar/foo",
        ) +
        CONFIG_TMPL_CUSTOM.format(
            name="label 2",
            linux="/images/pxeboot/vmlinuz root=hd:LABEL=root rhgb",
            initrd="/images/pxeboot/initrd.img",
        ) +
        "\n\n\n\n\n",
    ),
])
def test_grub2_iso(mocked_copy2, tmp_path, stage_module, test_data, expected_conf):
    treedir = tmp_path / "tree"
    treedir.mkdir(parents=True, exist_ok=True)
    efidir = treedir / "EFI/BOOT"
    confpath = efidir / "grub.cfg"

    # from fedora-ostree-bootiso-xz.json
    options = {
        "product": {
            "name": "Fedora",
            "version": "42"
        },
        "kernel": {
            "dir": "/images/pxeboot",
            "opts": [
                "inst.stage2=hd:LABEL=Fedora-42-Everything-x86_64"
            ]
        },
        "isolabel": "Fedora-42-Everything-x86_64",
        "architectures": [
            "X64"
        ],
        "vendor": "fedora"
    }
    options.update(test_data)

    stage_module.main(treedir, options)

    assert os.path.exists(confpath)
    assert confpath.read_text() == expected_conf
    assert mocked_copy2.call_args_list == [
        call("/boot/efi/EFI/fedora/shimx64.efi", os.fspath(efidir / "BOOTX64.EFI")),
        call("/boot/efi/EFI/fedora/mmx64.efi", os.fspath(efidir / "mmx64.efi")),
        call("/boot/efi/EFI/fedora/gcdx64.efi", os.fspath(efidir / "grubx64.efi")),
        call("/usr/share/grub/unicode.pf2", os.fspath(efidir / "fonts"))
    ]


@pytest.mark.parametrize("test_data,expected_err", [
    # bad
    (
        {}, ["'isolabel' is a required property", "'kernel' is a required property", "'product' is a required property"]
    ),
    (
        {
            "isolabel": "an-isolabel",
            "product": {
                "name": "a-name",
                "version": "a-version",
            },
            "kernel": {},
        }, ["'dir' is a required property"],
    ),
    (
        {
            "isolabel": "an-isolabel",
            "product": {},
            "kernel": {
                "dir": "/path/to",
            },
        }, ["'name' is a required property", "'version' is a required property"],
    ),
    # good
    (
        {
            "isolabel": "an-isolabel",
            "product": {
                "name": "a-name",
                "version": "a-version",
            },
            "kernel": {
                "dir": "/path/to",
            },
        }, "",
    ),
    # good + fips
    (
        {
            "isolabel": "an-isolabel",
            "product": {
                "name": "a-name",
                "version": "a-version",
            },
            "kernel": {
                "dir": "/path/to",
            },
            "fips": True,
        }, "",
    ),
    # good + default
    (
        {
            "isolabel": "an-isolabel",
            "product": {
                "name": "a-name",
                "version": "a-version",
            },
            "kernel": {
                "dir": "/path/to",
            },
            "config": {
                "default": 1,
            }
        }, "",
    ),
])
def test_schema_validation(stage_schema, test_data, expected_err):
    test_input = {
        "type": STAGE_NAME,
        "options": {},
    }
    test_input["options"].update(test_data)
    res = stage_schema.validate(test_input)

    if expected_err == "":
        assert res.valid is True, f"err: {[e.as_dict() for e in res.errors]}"
    else:
        assert res.valid is False
        err_msgs = sorted([e.as_dict()["message"] for e in res.errors])
        assert err_msgs == expected_err



#!/usr/bin/python3

import re

import pytest

from osbuild.testutil import (
    assert_jsonschema_error_contains,
    make_fake_tree,
)

STAGE_NAME = "org.osbuild.tuned"


@pytest.mark.parametrize("test_data,expected_err", [
    # bad
    ({"profiles": {}}, r"{} is not of type 'array'"),
    ({"profiles": ""}, r"'' is not of type 'array'"),
    ({"profiles": []}, r"(\[\] is too short|\[\] should be non-empty)"),
    ({"profiles": [0]}, r"0 is not of type 'string'"),
    ({"profiles": [""]}, r"('' is too short|'' should be non-empty)"),
    ({}, r"'profiles' is a required property"),
    # good
    ({"profiles": ["balanced"]}, ""),
    ({"profiles": ["balanced", "sap-hana"]}, ""),
])
def test_schema_validation(stage_schema, test_data, expected_err):
    test_input = {
        "type": STAGE_NAME,
        "options": {},
    }
    test_input["options"].update(test_data)
    res = stage_schema.validate(test_input)

    if expected_err == "":
        assert res.valid is True, f"err: {[e.as_dict() for e in res.errors]}"
    else:
        assert res.valid is False
        assert_jsonschema_error_contains(res, re.compile(expected_err), expected_num_errs=1)


@pytest.mark.parametrize("fake_tree,available_profiles", (
    ({}, []),
    # simple case with just installed profiles
    (
        {
            "/usr/lib/tuned/balanced/tuned.conf": "",
            "/usr/lib/tuned/sap/tuned.conf": ""
        },
        ["balanced", "sap"]
    ),
    # simple case with installed and custom profiles
    (
        {
            "/usr/lib/tuned/balanced/tuned.conf": "",
            "/usr/lib/tuned/sap/tuned.conf": "",
            "/etc/tuned/custom/tuned.conf": ""
        },
        ["balanced", "sap", "custom"]
    ),
    # Tuned 2.23.0+ with profiles under profiles/ directory
    # in which case we ignore profiles not under profiles/
    (
        {
            "/usr/lib/tuned/profiles/balanced/tuned.conf": "",
            "/usr/lib/tuned/sap/tuned.conf": ""
        },
        ["balanced"]
    ),
    (
        {
            "/usr/lib/tuned/profiles/balanced/tuned.conf": "",
            "/etc/tuned/sap/tuned.conf": ""
        },
        ["balanced", "sap"]
    ),
    (
        {
            "/usr/lib/tuned/profiles/balanced/tuned.conf": "",
            "/etc/tuned/profiles/sap/tuned.conf": ""
        },
        ["balanced", "sap"]
    ),
    (
        {
            "/etc/tuned/profiles/balanced/tuned.conf": "",
            "/etc/tuned/sap/tuned.conf": ""
        },
        ["balanced"]
    ),
    (
        {
            "/usr/lib/tuned/profiles/profile1/tuned.conf": "",
            "/usr/lib/tuned/profiles/profile2/tuned.conf": "",
            "/usr/lib/tuned/profile3/tuned.conf": "",
            "/etc/tuned/profiles/profile11/tuned.conf": "",
            "/etc/tuned/profiles/profile12/tuned.conf": "",
            "/etc/tuned/profile13/tuned.conf": "",
        },
        ["profile1", "profile2", "profile11", "profile12"]
    ),
))
def test_tunedprofilesdb__load_available_profiles(tmp_path, stage_module, fake_tree, available_profiles):
    make_fake_tree(tmp_path, fake_tree)
    # pylint: disable=protected-access
    assert sorted(stage_module.TunedProfilesDB._load_available_profiles(tmp_path)) == sorted(available_profiles)


def test_tuned_happy(tmp_path, stage_module):
    make_fake_tree(tmp_path, {
        "/usr/lib/tuned/balanced/tuned.conf": "",
        "/usr/lib/tuned/sap/tuned.conf": "",
        "/etc/tuned/custom/tuned.conf": "",
    })

    options = {
        "profiles": ["balanced", "sap", "custom"]
    }

    assert stage_module.main(tmp_path.as_posix(), options) == 0

    active_profile_file = tmp_path / "etc/tuned/active_profile"
    assert active_profile_file.is_file()
    assert active_profile_file.read_text() == "balanced sap custom\n"

    profile_mode_file = tmp_path / "etc/tuned/profile_mode"
    assert profile_mode_file.is_file()
    assert profile_mode_file.read_text() == "manual\n"


def test_tuned_unhappy(tmp_path, stage_module):
    make_fake_tree(tmp_path, {
        "/usr/lib/tuned/balanced/tuned.conf": "",
        "/usr/lib/tuned/sap/tuned.conf": "",
        "/etc/tuned/custom/tuned.conf": "",
    })

    options = {
        "profiles": ["balanced", "non-existing"]
    }

    try:
        stage_module.main(tmp_path.as_posix(), options)
    except ValueError as e:
        assert "non-existing" in str(e)
    else:
        assert False, "Exception not raised"

    active_profile_file = tmp_path / "etc/tuned/active_profile"
    assert not active_profile_file.exists()

    profile_mode_file = tmp_path / "etc/tuned/profile_mode"
    assert not profile_mode_file.exists()



#!/usr/bin/python3

import re

import pytest

from osbuild.testutil import assert_jsonschema_error_contains

STAGE_NAME = "org.osbuild.grub2.d"


@pytest.mark.parametrize("test_data,expected_err", [
    # good - tree:// paths
    ({"config": {"serial": "serial --speed=115200"}, "path": "tree:///boot/grub2/console.cfg"}, ""),
    ({"config": {"terminal_input": ["serial", "console"]}, "path": "tree:///boot/grub2/console.cfg"}, ""),
    ({"config": {"terminal_output": ["serial", "console"]}, "path": "tree:///boot/grub2/console.cfg"}, ""),
    ({"config": {
        "serial": "serial --speed=115200",
        "terminal_input": ["serial", "console"],
        "terminal_output": ["serial", "console"],
    }, "path": "tree:///boot/grub2/console.cfg"}, ""),
    # good - mount:// paths
    ({"config": {"serial": "serial --speed=115200"}, "path": "mount://root/boot/grub2/custom.cfg"}, ""),
    ({"config": {"serial": "serial --speed=115200"}, "path": "mount://root/etc/grub.d/99-serial.cfg"}, ""),
    # bad - missing required config
    ({"path": "tree:///boot/grub2/console.cfg"}, "'config' is a required property"),
    # bad - missing required path
    ({"config": {"serial": "serial --speed=115200"}}, "'path' is a required property"),
    # bad - empty config
    ({"config": {}, "path": "tree:///boot/grub2/console.cfg"},
     re.compile(r"{} should be non-empty|{} does not have enough properties")),
    # bad - unknown property in config
    ({"config": {"unknown": "value"}, "path": "tree:///boot/grub2/console.cfg"}, "Additional properties are not allowed"),
    # bad - wrong types
    ({"config": {"serial": 123}, "path": "tree:///boot/grub2/console.cfg"}, "123 is not of type 'string'"),
    ({"config": {"terminal_input": "not-an-array"}, "path": "tree:///boot/grub2/console.cfg"},
     "'not-an-array' is not of type 'array'"),
    ({"config": {"terminal_output": "not-an-array"}, "path": "tree:///boot/grub2/console.cfg"},
     "'not-an-array' is not of type 'array'"),
    ({"config": {"terminal_input": [123]}, "path": "tree:///boot/grub2/console.cfg"},
     "123 is not of type 'string'"),
    # bad - unknown top-level option
    ({"config": {"serial": "serial"}, "unknown": "value", "path": "tree:///x"}, "Additional properties are not allowed"),
    # bad - bare path (no scheme)
    ({"config": {"serial": "serial"}, "path": "boot/grub2/console.cfg"}, "is not valid under any of the given schemas"),
    ({"config": {"serial": "serial"}, "path": "/boot/grub2/console.cfg"}, "is not valid under any of the given schemas"),
])
def test_schema_validation(stage_schema, test_data, expected_err):
    test_input = {
        "type": STAGE_NAME,
        "options": {},
    }
    test_input["options"].update(test_data)
    res = stage_schema.validate(test_input)

    if expected_err == "":
        assert res.valid is True, f"err: {[e.as_dict() for e in res.errors]}"
    else:
        assert res.valid is False
        assert_jsonschema_error_contains(res, expected_err)


@pytest.mark.parametrize("config,expected_lines", [
    # empty config produces empty file
    ({}, []),
    # serial only
    (
        {"serial": "serial --speed=115200"},
        ["serial --speed=115200"],
    ),
    # terminal_input only
    (
        {"terminal_input": ["serial", "console"]},
        ["terminal_input serial console"],
    ),
    # terminal_output only
    (
        {"terminal_output": ["serial"]},
        ["terminal_output serial"],
    ),
    # all three
    (
        {
            "serial": "serial --speed=115200 --unit=0 --word=8 --parity=no --stop=1",
            "terminal_input": ["serial", "console"],
            "terminal_output": ["serial", "console"],
        },
        [
            "serial --speed=115200 --unit=0 --word=8 --parity=no --stop=1",
            "terminal_input serial console",
            "terminal_output serial console",
        ],
    ),
])
def test_generate_grub2_dropin(tmp_path, stage_module, config, expected_lines):
    dropin_path = tmp_path / "test.cfg"
    stage_module.generate_grub2_dropin(config, dropin_path)

    content = dropin_path.read_text(encoding="utf8")
    if expected_lines:
        assert content == "\n".join(expected_lines) + "\n"
    else:
        assert content == ""


def test_main_tree_path(tmp_path, stage_module):
    """main() writes to the specified tree:// location."""
    args = {"tree": str(tmp_path)}
    options = {
        "path": "tree:///boot/grub2/console.cfg",
        "config": {
            "serial": "serial --speed=115200",
            "terminal_input": ["serial", "console"],
            "terminal_output": ["serial", "console"],
        },
    }
    stage_module.main(args, options)

    cfg_path = tmp_path / "boot" / "grub2" / "console.cfg"
    assert cfg_path.exists()
    content = cfg_path.read_text(encoding="utf8")
    assert "serial --speed=115200\n" in content
    assert "terminal_input serial console\n" in content
    assert "terminal_output serial console\n" in content


def test_main_creates_parent_dirs(tmp_path, stage_module):
    """main() creates parent directories if they don't exist."""
    args = {"tree": str(tmp_path)}
    options = {
        "path": "tree:///boot/grub2/console.cfg",
        "config": {
            "serial": "serial --speed=115200",
        },
    }
    stage_module.main(args, options)

    cfg_path = tmp_path / "boot" / "grub2" / "console.cfg"
    assert cfg_path.exists()



#!/usr/bin/python3

import os
import subprocess as sp
import textwrap

import pytest

STAGE_NAME = "org.osbuild.coreos.live-artifacts.mono"


def test_get_os_features(tmp_path, stage_module):
    cfg_path = tmp_path / "usr/share/coreos-installer/example-config.yaml"
    cfg_path.parent.mkdir(parents=True)
    cfg_path.write_text(textwrap.dedent("""\
    # Fedora CoreOS stream
    stream: name
    # Manually specify the image URL
    image-url: URL
    """))
    features = stage_module.get_os_features(tmp_path)
    assert {
        "installer-config": True,
        "installer-config-directives": {
            "stream": True,
            "image-url": True,
        },
        "live-initrd-network": True,
    } == features


@pytest.mark.parametrize("test_case", [
    {"orig_size": 0, "exp_size": 0},
    {"orig_size": 1, "exp_size": 4},
    {"orig_size": 2, "exp_size": 4},
    {"orig_size": 3, "exp_size": 4},
    {"orig_size": 4, "exp_size": 4},
    {"orig_size": 10938, "exp_size": 10940},
])
def test_align_initrd(tmp_path, stage_module, test_case):
    dest_file = tmp_path / "rootfs.img"
    dest_file.touch()

    os.truncate(dest_file, test_case["orig_size"])
    stage_module.align_initrd_for_uncompressed_append(dest_file.open(mode="ab"))
    assert dest_file.stat().st_size == test_case["exp_size"]

    with dest_file.open(mode="rb") as fp:
        fp.seek(test_case["orig_size"])
        padding_size = test_case["exp_size"] - test_case["orig_size"]
        padding = fp.read(padding_size)
    assert padding == b'\0' * padding_size


@pytest.mark.parametrize("test_case", [
    {"treefiles": ["file_b", "file_c", "file_1"], "compress": False},
    {"treefiles": ["file_b", "file_c", "root.squashfs", "file_1"], "compress": False},
    {"treefiles": ["file_b", "file_c", "file_1"], "compress": True},
    {"treefiles": ["file_b", "file_c", "root.squashfs", "file_1"], "compress": True},
    {"treefiles": ["file_b", "file_c", "root.erofs", "file_1"], "compress": True},
])
def test_extend_initramfs(tmp_path, stage_module, test_case):
    img_path = tmp_path / "img"
    img_path.mkdir()
    dst = img_path / "dst.img"
    dst.touch()

    tree = tmp_path / "tree"
    tree.mkdir()
    treefiles = test_case["treefiles"]
    for fname in treefiles:
        (tree / fname).touch()

    stage_module.extend_initramfs(dst, tree, compress=test_case["compress"])

    if test_case["compress"]:
        # if we compressed, append .gz to the filename and let gunzip remove it when decompressing
        gz_name = str(dst) + ".gz"
        dst.rename(gz_name)
        sp.run(["gunzip", gz_name], check=True)

    output = sp.check_output(["cpio", "-i", "-t", "--file", dst])
    files_in_archive = output.decode().strip().split("\n")

    # check that all the elements match, regardless of order
    assert sorted(treefiles) == sorted(files_in_archive)
    if "root.squashfs" in treefiles:
        # root.squashfs must be first
        assert files_in_archive[0] == "root.squashfs"
    if "root.erofs" in treefiles:
        # root.erofs must be first
        assert files_in_archive[0] == "root.erofs"


def test_make_stream_hash(tmp_path, stage_module):
    src = tmp_path / "src"
    dst = tmp_path / "dst"

    with src.open(mode="wb") as src_file:
        # repeat enough times to require multiple buffers (bufsize == 2097152)
        src_file.write(b"TNdUoOTJgTAfeNYl9anmDpo8CxphoMaAoEESZWdwj45RZXNTHv7Fwsj5" * 100_000)

    stage_module.make_stream_hash(src, dst)
    with dst.open("r") as dst_file:
        stream_hash = dst_file.read()

    hashes = [
        "6f50ae3331d2d010465b42292e82e53dcc24548ef317eee80e71f657f56995d3",
        "d4170e95e324afa6b687a8cb45144e1edf75e3ac6b2c63e12dc31884b0bbfd01",
        "2f3599a36686b38dea4937c8c2dccd00574d03e0cba07e03e791f07c8f06bff1",
    ]
    assert stream_hash.strip().split("\n") == [
        "stream-hash sha256 2097152",
        *hashes,
    ]



"""userspace /dev device manager (udev) utilities"""

import contextlib
import pathlib

# The default lock dir to use
LOCKDIR = "/run/osbuild/locks/udev"


class UdevInhibitor:
    """
    Inhibit execution of certain udev rules for block devices

    This is the osbuild side of the custom mechanism that
    allows us to inhibit certain udev rules for block devices.

    For each device a lock file is created in a well known
    directory (LOCKDIR). A custom udev rule set[1] checks
    for the said lock file and inhibits other udev rules from
    being executed.
    See the aforementioned rules file for more information.

    [1] 10-osbuild-inhibitor.rules
    """

    def __init__(self, path: pathlib.Path):
        self.path = path
        path.parent.mkdir(parents=True, exist_ok=True)

    def inhibit(self) -> None:
        self.path.touch()

    def release(self) -> None:
        with contextlib.suppress(FileNotFoundError):
            self.path.unlink()

    @property
    def active(self) -> bool:
        return self.path.exists()

    def __str__(self):
        return f"UdevInhibtor at '{self.path}'"

    @classmethod
    def for_dm_name(cls, name: str, lockdir=LOCKDIR):
        """Inhibit a Device Mapper device with the given name"""
        path = pathlib.Path(lockdir, f"dm-{name}")
        ib = cls(path)
        ib.inhibit()
        return ib

    @classmethod
    def for_device(cls, major: int, minor: int, lockdir=LOCKDIR):
        """Inhibit a device given its major and minor number"""
        path = pathlib.Path(lockdir, f"device-{major}:{minor}")
        ib = cls(path)
        ib.inhibit()
        return ib



import os
from datetime import datetime
from typing import Dict, List, Optional, Union
from uuid import uuid4

import osbuild
import osbuild.util.sbom.model as sbom_model
import osbuild.util.sbom.spdx2 as spdx2

try:
    from license_expression import ExpressionError, get_spdx_licensing
except ImportError:
    get_spdx_licensing = None
    ExpressionError = None


class SpdxLicenseExpressionCreator:
    """
    Class for creating SPDX license expressions from license strings.

    This class uses the license-expression package to parse license strings and convert them to SPDX license, if
    possible.

    The class object also keeps track of all extracted licensing information objects that were created during the
    conversion process. The extracted licensing information objects are stored in a dictionary, where the key is the
    license reference ID and the value is the ExtractedLicensingInfo object.
    """

    def __init__(self, license_index_location=None):
        self._extracted_license_infos: Dict[str, spdx2.ExtractedLicensingInfo] = {}
        self._spdx_licensing = None

        if get_spdx_licensing:
            if license_index_location:
                self._spdx_licensing = get_spdx_licensing(license_index_location)
            else:
                self._spdx_licensing = get_spdx_licensing()
        elif license_index_location:
            raise ValueError("The license-expression package is not available. "
                             "Specify the license index location has no effect.")

    def _to_extracted_license_info(self, license_str: str) -> spdx2.ExtractedLicensingInfo:
        eli = spdx2.ExtractedLicensingInfo(license_str)
        return self._extracted_license_infos.setdefault(eli.license_ref_id, eli)

    def ensure_license_expression(self, license_str: str) -> Union[str, spdx2.ExtractedLicensingInfo]:
        """
        Convert a license string to a valid SPDX license expression or wrap it in an ExtractedLicensingInfo object.

        This function uses the license-expression package to parse the license string and convert it to an SPDX license
        expression. If the license string can't be parsed and converted to an SPDX license expression, it is wrapped in an
        ExtractedLicensingInfo object.

        If the license-expression package is not available, the license string is always wrapped in an
        ExtractedLicensingInfo object.

        License strings that are already SPDX license ref IDs are returned as is.
        """
        if license_str.startswith("LicenseRef-"):
            # The license string is already an SPDX license ref ID.
            return license_str

        if self._spdx_licensing is None:
            return self._to_extracted_license_info(license_str)

        try:
            return str(self._spdx_licensing.parse(license_str, validate=True, strict=True))
        except ExpressionError:
            return self._to_extracted_license_info(license_str)

    def extracted_license_infos(self) -> List[spdx2.ExtractedLicensingInfo]:
        """
        Return a list of all extracted licensing information objects that were created during the conversion process.
        """
        return list(self._extracted_license_infos.values())


def spdx2_checksum_algorithm(algorithm: sbom_model.ChecksumAlgorithm) -> spdx2.ChecksumAlgorithm:
    if algorithm == sbom_model.ChecksumAlgorithm.SHA1:
        return spdx2.ChecksumAlgorithm.SHA1
    if algorithm == sbom_model.ChecksumAlgorithm.SHA224:
        return spdx2.ChecksumAlgorithm.SHA224
    if algorithm == sbom_model.ChecksumAlgorithm.SHA256:
        return spdx2.ChecksumAlgorithm.SHA256
    if algorithm == sbom_model.ChecksumAlgorithm.SHA384:
        return spdx2.ChecksumAlgorithm.SHA384
    if algorithm == sbom_model.ChecksumAlgorithm.SHA512:
        return spdx2.ChecksumAlgorithm.SHA512
    if algorithm == sbom_model.ChecksumAlgorithm.MD5:
        return spdx2.ChecksumAlgorithm.MD5
    raise ValueError(f"Unknown checksum algorithm: {algorithm}")


def create_spdx2_document():
    tool = f"osbuild-{osbuild.__version__}"
    doc_name = f"sbom-by-{tool}"

    ci = spdx2.CreationInfo(
        spdx_version="SPDX-2.3",
        spdx_id="SPDXRef-DOCUMENT",
        name=doc_name,
        data_license="CC0-1.0",
        document_namespace=f"https://osbuild.org/spdxdocs/{doc_name}-{uuid4()}",
        creators=[spdx2.Creator(spdx2.CreatorType.TOOL, tool)],
        created=datetime.now(),
    )
    doc = spdx2.Document(ci)

    return doc


def sbom_pkgset_to_spdx2_doc(
        pkgset: List[sbom_model.BasePackage],
        license_index_location: Optional[os.PathLike] = None) -> spdx2.Document:
    doc = create_spdx2_document()
    relationships = []
    license_expr_creator = SpdxLicenseExpressionCreator(license_index_location)

    for pkg in pkgset:

        download_location: Union[str, spdx2.NoAssertionValue] = spdx2.NoAssertionValue()
        if pkg.download_url:
            download_location = pkg.download_url

        license_declared = license_expr_creator.ensure_license_expression(pkg.license_declared)

        p = spdx2.Package(
            spdx_id=f"SPDXRef-{pkg.uuid()}",
            name=pkg.name,
            download_location=download_location,
            version=pkg.version,
            files_analyzed=False,
            license_declared=license_declared,
            external_references=[
                spdx2.ExternalPackageRef(
                    category=spdx2.ExternalPackageRefCategory.PACKAGE_MANAGER,
                    reference_type="purl",
                    locator=pkg.purl(),
                )
            ]
        )

        if pkg.homepage:
            p.homepage = pkg.homepage

        if pkg.summary:
            p.summary = pkg.summary

        if pkg.description:
            p.description = pkg.description

        if pkg.source_info():
            p.source_info = pkg.source_info()

        for hash_type, hash_value in pkg.checksums.items():
            p.checksums.append(
                spdx2.Checksum(
                    algorithm=spdx2_checksum_algorithm(hash_type),
                    value=hash_value,
                )
            )

        if pkg.build_date:
            p.built_date = pkg.build_date

        doc.packages.append(p)

        relationships.append(
            spdx2.Relationship(
                spdx_element_id=doc.creation_info.spdx_id,
                relationship_type=spdx2.RelationshipType.DESCRIBES,
                related_spdx_element_id=p.spdx_id,
            )
        )

        for dep in sorted(pkg.depends_on, key=lambda x: x.uuid()):
            relationships.append(
                spdx2.Relationship(
                    spdx_element_id=p.spdx_id,
                    relationship_type=spdx2.RelationshipType.DEPENDS_ON,
                    related_spdx_element_id=f"SPDXRef-{dep.uuid()}",
                )
            )

        for optional_dep in sorted(pkg.optional_depends_on, key=lambda x: x.uuid()):
            relationships.append(
                spdx2.Relationship(
                    spdx_element_id=f"SPDXRef-{optional_dep.uuid()}",
                    relationship_type=spdx2.RelationshipType.OPTIONAL_DEPENDENCY_OF,
                    related_spdx_element_id=p.spdx_id,
                )
            )

    doc.relationships = relationships

    extracted_license_infos = license_expr_creator.extracted_license_infos()
    if len(extracted_license_infos) > 0:
        doc.extracted_licensing_infos = extracted_license_infos

    return doc



#!/usr/bin/python3

import os
import pathlib
import tempfile

import pytest

from osbuild.testutil import has_executable
from osbuild.testutil.net import http_serve_directory, https_serve_directory
from osbuild.util import ostree

SOURCES_NAME = "org.osbuild.ostree"


@pytest.mark.skipif(not has_executable("ostree"), reason="need ostree")
def test_ostree_source_not_exists(tmp_path, sources_service):
    checksum = "sha256:1111111111111111111111111111111111111111111111111111111111111111"
    sources_service.setup({"cache": tmp_path, "options": {}})
    assert not sources_service.exists(checksum, None)


@pytest.mark.skipif(not has_executable("ostree"), reason="need ostree")
def test_ostree_source_exists(tmp_path, sources_service):
    sources_service.setup({"cache": tmp_path, "options": {}})
    root = tmp_path / "org.osbuild.ostree" / "repo"
    commit = make_repo(root)
    assert sources_service.exists("sha256:" + commit, None)


def make_test_sources(proto, port, fake_commit, **secrets):
    sources = {
        fake_commit: {
            "remote": {
                "url": f"{proto}://localhost:{port}",
            }
        }
    }
    if secrets:
        sources[fake_commit]["remote"]["secrets"] = secrets
    return sources


def make_repo(root):
    with tempfile.TemporaryDirectory() as empty_tmpdir:
        ostree.cli("init", f"--repo={root}")
        return ostree.cli("commit", f"--repo={root}", "--orphan", empty_tmpdir).stdout.rstrip()


@pytest.mark.skipif(not has_executable("ostree"), reason="need ostree")
def test_ostree_pull_plain(tmp_path, sources_service):
    fake_httpd_root = tmp_path / "fake-httpd-root"
    fake_httpd_root.mkdir(exist_ok=True)
    fake_commit = make_repo(fake_httpd_root)

    with http_serve_directory(fake_httpd_root) as httpd:
        test_sources = make_test_sources("http", httpd.server_port, fake_commit)
        sources_service.setup({"cache": tmp_path, "options": {}})
        sources_service.fetch_all(test_sources)
        assert sources_service.exists("sha256:" + fake_commit, None)


@pytest.mark.skipif(not has_executable("ostree"), reason="need ostree")
def test_ostree_pull_plain_mtls(tmp_path, sources_service, monkeypatch):
    fake_httpd_root = tmp_path / "fake-httpd-root"
    fake_httpd_root.mkdir(exist_ok=True)
    fake_commit = make_repo(fake_httpd_root)

    cert_dir = pathlib.Path(__file__).parent.parent.parent / "test" / "data" / "certs"
    cert1 = cert_dir / "cert1.pem"
    assert cert1.exists()
    key1 = cert_dir / "key1.pem"
    assert key1.exists()

    with https_serve_directory(fake_httpd_root, cert1, key1) as httpd:
        monkeypatch.setenv("OSBUILD_SOURCES_OSTREE_INSECURE", "1")
        test_sources = make_test_sources("https", httpd.server_port, fake_commit, name="org.osbuild.mtls")
        sources_service.setup({"cache": tmp_path, "options": {}})
        sources_service.fetch_all(test_sources)
        assert sources_service.exists("sha256:" + fake_commit, None)


@pytest.mark.skipif(not has_executable("ostree"), reason="need ostree")
def test_ostree_copy_all(tmp_path, sources_service):
    fake_httpd_root = tmp_path / "fake-httpd-root"
    fake_httpd_root.mkdir(exist_ok=True)
    fake_commit = make_repo(fake_httpd_root)
    src_cache = tmp_path / "src-cache"
    dst_cache = tmp_path / "dst-cache"

    os.mkdir(src_cache)
    os.mkdir(dst_cache)

    with http_serve_directory(fake_httpd_root) as httpd:
        test_sources = make_test_sources("http", httpd.server_port, fake_commit)
        sources_service.setup({"cache": src_cache, "options": {}})
        sources_service.fetch_all(test_sources)
        sources_service.copy_all(test_sources, dst_cache)
        assert sources_service.exists("sha256:" + fake_commit, None)
        # ensure the commit exists in the dst_cache repo
        ostree.show(dst_cache / sources_service.content_type / "repo", "sha256:" + fake_commit)



#!/usr/bin/python3

import os.path
from unittest.mock import call, patch

import pytest

STAGE_NAME = "org.osbuild.isolinux"

CONFIG_PART_1 = """
default vesamenu.c32
timeout 600

display boot.msg

# Clear the screen when exiting the menu, instead of leaving the menu displayed.
# For vesamenu, this means the graphical background is still displayed without
# the menu itself for as long as the screen remains in graphics mode.
menu clear
menu background splash.png
menu title Fedora 42
menu vshift 8
menu rows 18
menu margin 8
#menu hidden
menu helpmsgrow 15
menu tabmsgrow 13

# Border Area
menu color border * #00000000 #00000000 none

# Selected item
menu color sel 0 #ffffffff #00000000 none

# Title bar
menu color title 0 #ff7ba3d0 #00000000 none

# Press [Tab] message
menu color tabmsg 0 #ff3a6496 #00000000 none

# Unselected menu item
menu color unsel 0 #84b8ffff #00000000 none

# Selected hotkey
menu color hotsel 0 #84b8ffff #00000000 none

# Unselected hotkey
menu color hotkey 0 #ffffffff #00000000 none

# Help text
menu color help 0 #ffffffff #00000000 none

# A scrollbar of some type? Not sure.
menu color scrollbar 0 #ffffffff #ff355594 none

# Timeout msg
menu color timeout 0 #ffffffff #00000000 none
menu color timeout_msg 0 #ffffffff #00000000 none

# Command prompt text
menu color cmdmark 0 #84b8ffff #00000000 none
menu color cmdline 0 #ffffffff #00000000 none

# Do not display the actual menu unless the user presses a key. All that is displayed is a timeout message.

menu tabmsg Press Tab for full configuration options on menu items.

menu separator # insert an empty line
menu separator # insert an empty line

label linux
  menu label ^Install Fedora 42
  kernel vmlinuz
  append initrd=initrd.img inst.stage2=hd:LABEL=Fedora-42-Everything-x86_64 quiet

label check
  menu label Test this ^media & install Fedora 42
  menu default
  kernel vmlinuz
  append initrd=initrd.img inst.stage2=hd:LABEL=Fedora-42-Everything-x86_64 rd.live.check quiet

"""

CONFIG_PART_2 = """
menu separator # insert an empty line

# utilities submenu
menu begin ^Troubleshooting
  menu title Troubleshooting Fedora 42

label basic
  menu indent count 5
  menu label Install using ^basic graphics mode
  text help
	Try this option out if you're having trouble installing
	Fedora 42.
  endtext
  kernel vmlinuz
  append initrd=initrd.img inst.stage2=hd:LABEL=Fedora-42-Everything-x86_64 nomodeset quiet
label rescue
  menu indent count 5
  menu label ^Rescue a Fedora system
  text help
	If the system will not boot, this lets you access files
	and edit config files to try to get it booting again.
  endtext
  kernel vmlinuz
  append initrd=initrd.img inst.stage2=hd:LABEL=Fedora-42-Everything-x86_64 inst.rescue quiet
label memtest
  menu label Run a ^memory test
  text help
	If your system is having issues, a problem with your
	system's memory may be the cause. Use this utility to
	see if the memory is working correctly.
  endtext
  kernel memtest

menu separator # insert an empty line

label local
  menu label Boot from ^local drive
  localboot 0xffff

menu separator # insert an empty line
menu separator # insert an empty line

label returntomain
  menu label Return to ^main menu
  menu exit

"""

CONFIG_FIPS = """
label fips
  menu label ^Install Fedora 42 in FIPS mode
  kernel vmlinuz
  append initrd=initrd.img inst.stage2=hd:LABEL=Fedora-42-Everything-x86_64 quiet fips=1

"""


@patch("os.link")
@patch("os.chmod")
@patch("shutil.copyfile")
@pytest.mark.parametrize("test_data,expected_conf", [
    # default
    ({}, CONFIG_PART_1 + CONFIG_PART_2),
    # fips menu enable
    ({"fips": True}, CONFIG_PART_1 + CONFIG_FIPS + CONFIG_PART_2)
])
def test_isolinux(mocked_copyfile, mocked_chmod, mocked_link, tmp_path, stage_module, test_data, expected_conf):
    treedir = tmp_path / "tree"
    treedir.mkdir(parents=True, exist_ok=True)
    datadir = tmp_path / "data"
    confpath = treedir / "isolinux/isolinux.cfg"

    # from fedora-ostree-bootiso-xz.json
    options = {
        "product": {
            "name": "Fedora",
            "version": "42"
        },
        "kernel": {
            "dir": "/images/pxeboot",
            "opts": [
                "inst.stage2=hd:LABEL=Fedora-42-Everything-x86_64"
            ]
        },
    }
    options.update(test_data)

    inputs = {
        "data": {
            "path": datadir
        }
    }

    stage_module.main(treedir, inputs, options)

    assert os.path.exists(confpath)
    assert confpath.read_text() == expected_conf
    assert mocked_copyfile.call_args_list == [
        call(os.fspath(datadir / "usr/share/anaconda/boot/syslinux-splash.png"),
             os.fspath(treedir / "isolinux/splash.png")),
        call(os.fspath(datadir / "usr/share/syslinux/isolinux.bin"),
             os.fspath(treedir / "isolinux/isolinux.bin")),
        call(os.fspath(datadir / "usr/share/syslinux/ldlinux.c32"),
             os.fspath(treedir / "isolinux/ldlinux.c32")),
        call(os.fspath(datadir / "usr/share/syslinux/libcom32.c32"),
             os.fspath(treedir / "isolinux/libcom32.c32")),
        call(os.fspath(datadir / "usr/share/syslinux/libutil.c32"),
             os.fspath(treedir / "isolinux/libutil.c32")),
        call(os.fspath(datadir / "usr/share/syslinux/vesamenu.c32"),
             os.fspath(treedir / "isolinux/vesamenu.c32")),
    ]
    assert mocked_chmod.call_args_list == [
        call(os.fspath(treedir / "isolinux/isolinux.bin"), 0o755),
        call(os.fspath(treedir / "isolinux/ldlinux.c32"), 0o755),
        call(os.fspath(treedir / "isolinux/libcom32.c32"), 0o755),
        call(os.fspath(treedir / "isolinux/libutil.c32"), 0o755),
        call(os.fspath(treedir / "isolinux/vesamenu.c32"), 0o755),
    ]
    assert mocked_link.call_args_list == [
        call(os.fspath(treedir / "images/pxeboot/vmlinuz"),
             os.fspath(treedir / "isolinux/vmlinuz")),
        call(os.fspath(treedir / "images/pxeboot/initrd.img"),
             os.fspath(treedir / "isolinux/initrd.img"))
    ]


@pytest.mark.parametrize("test_data,expected_err", [
    # bad
    (
        {}, ["'kernel' is a required property", "'product' is a required property"]
    ),
    (
        {
            "product": {
                "name": "a-name",
                "version": "a-version",
            },
            "kernel": {},
        }, ["'dir' is a required property"],
    ),
    (
        {
            "product": {},
            "kernel": {
                "dir": "/path/to",
            },
        }, ["'name' is a required property", "'version' is a required property"],
    ),
    # good
    (
        {
            "product": {
                "name": "a-name",
                "version": "a-version",
            },
            "kernel": {
                "dir": "/path/to",
            },
        }, "",
    ),
    # good + fips
    (
        {
            "product": {
                "name": "a-name",
                "version": "a-version",
            },
            "kernel": {
                "dir": "/path/to",
            },
            "fips": True,
        }, "",
    ),
])
def test_schema_validation(stage_schema, test_data, expected_err):
    test_input = {
        "type": STAGE_NAME,
        "options": {},
    }
    test_input["options"].update(test_data)
    res = stage_schema.validate(test_input)

    if expected_err == "":
        assert res.valid is True, f"err: {[e.as_dict() for e in res.errors]}"
    else:
        assert res.valid is False
        err_msgs = sorted([e.as_dict()["message"] for e in res.errors])
        assert err_msgs == expected_err



#!/usr/bin/python3

import os.path
import zipfile

import pytest

from osbuild import testutil
from osbuild.testutil import has_executable, make_fake_input_tree

STAGE_NAME = "org.osbuild.zip"


@pytest.mark.parametrize("test_data,expected_err", [
    # bad
    ({}, "'filename' is a required property"),
    ({"filename": "x", "level": 99}, "99 is greater than the maximum of 9"),
    # good
    ({"filename": "image.zip"}, ""),
])
def test_schema_validation_zip(stage_schema, test_data, expected_err):
    test_input = {
        "type": STAGE_NAME,
        "options": {
        }
    }
    test_input["options"].update(test_data)
    res = stage_schema.validate(test_input)

    if expected_err == "":
        assert res.valid is True, f"err: {[e.as_dict() for e in res.errors]}"
    else:
        assert res.valid is False
        testutil.assert_jsonschema_error_contains(res, expected_err, expected_num_errs=1)


@pytest.mark.parametrize("only_include, expected", [
    ([], ["file-in-root.txt", "subdir/", "subdir/subdir.txt", "subdir2/", "subdir2/subdir2.txt"]),
    (["subdir/*"], ["subdir/", "subdir/subdir.txt"]),
    (["subdir/*", "subdir2/*"], ["subdir/", "subdir/subdir.txt", "subdir2/", "subdir2/subdir2.txt"]),
])
@pytest.mark.skipif(not has_executable("zip"), reason="no zip executable")
def test_zip_integration(tmp_path, stage_module, only_include, expected):
    fake_input_tree = make_fake_input_tree(tmp_path, {
        "/file-in-root.txt": "other content",
        "/subdir/subdir.txt": "subdir content",
        "/subdir2/subdir2.txt": "subdir2 content",
    })
    inputs = {
        "tree": {
            "path": fake_input_tree,
        }
    }
    options = {
        "filename": "archive.zip",
    }
    if only_include:
        options["include"] = only_include
    output_dir = tmp_path
    stage_module.main(inputs, output_dir, options)
    expected_zip_path = output_dir / "archive.zip"
    assert os.path.exists(expected_zip_path)
    with zipfile.ZipFile(expected_zip_path, "r") as zfp:
        assert sorted(zfp.namelist()) == sorted(expected)
